import jarray
import struct
from util import Helper
from visad import *
from visad.util import DataUtility
from java.io import FileInputStream
from java.util import Properties
from visad.python.JPythonMethods import *
from java.lang import Double, Float, Math, System, Object
from java.net import URL
from visad.data.units import *

from subsetImage import Subset

import hydra_globals
from hydra_globals import hydra_path

false = 0
true  = 1
    
#---------------------------------------------------------------------------------    
class AbstractNameAssociationTable:
  def __init__(self, filename):
    self.filename = filename #- this table's filename
    fis = FileInputStream(hydra_path+"/multispectral/properties/"+filename)
    p   = Properties()
    p.load(fis)
    self.p = p
    
    self.instrumentNameAttribute = p.getProperty("instrumentNameAttribute")
    self.instrumentName          = p.getProperty("instrumentName")
    self.observation             = p.getProperty("observation")
    self.observationChannels     = p.getProperty("observationChannels")
    self.channelIntervals        = p.getProperty("obsChannelInterval")
    self.channelIndex            = p.getProperty("channelIndex")
    if self.channelIndex == "":
       self.channelIndex=None
    self.obsElement              = p.getProperty("obsElement")
    self.obsLine                 = p.getProperty("obsLine")
    self.obsLevel                = p.getProperty("obsLevels")
    self.obsLatitude             = p.getProperty("obsLatitude")
    self.obsLongitude            = p.getProperty("obsLongitude")
    self.segmentBounds           = p.getProperty("segmentBounds")
    self.obsFOVangle             = p.getProperty("obsFOVangle")
    self.obsTimeOfDay            = p.getProperty("obsTimeOfDay")
    self.geoLonElement           = p.getProperty("geoLonElement")
    self.geoLonLine              = p.getProperty("geoLonLine")
    self.geoLatElement           = p.getProperty("geoLatElement")
    self.geoLatLine              = p.getProperty("geoLatLine")
    self.range_long_name         = p.getProperty("range_long_name")
    self.range_units             = p.getProperty("range_units")
    self.range_valid_min         = p.getProperty("range_valid_min")
    self.range_valid_max         = p.getProperty("range_valid_max")
    self.range_valid_type        = p.getProperty("range_valid_type")
    self.range_missing           = p.getProperty("range_missing")
    self.solarZenith             = p.getProperty("solarzenith")
    
    
  #- verify this name translation table with file opened by reader
  #
  def verifyTable(self, reader):
    names = []
    
    a = self.observation.split(",")
    for i in xrange(len(a)):
      b = a[i].split("-")  # look for minus signs (differences of existing fields as for 
                           # AIRS Level 2 Difference of Surface Temperature and Air Temperature)
      for j in xrange(len(b)):
        # For checking, name only needs to appear once in obsElement string
        if b[j] not in names:
          names.append(b[j])

    if self.channelIndex != None:
      a = self.channelIndex.split(",")
      for i in xrange(len(a)):
        names.append(a[i])
    
    a = self.obsElement.split(",")
    for i in xrange(len(a)):
      names.append(a[i])
      
    a = self.obsLine.split(",")
    for i in xrange(len(a)):
      names.append(a[i])

    if self.instrumentNameAttribute != None:
      names.append(self.instrumentNameAttribute)

    names.append(self.obsLatitude)
    names.append(self.obsLongitude)
     
    for i in xrange(len(names)):
      if names[i] != None:
        ok = reader.verifyName(names[i])
        if (ok == false): return false
        
    return true
#---------------------------------------------------------------------------
#- Find a local/remote reader for this <filename, url>; return an
#  instance of MultiSpectralData object which has methods for accessing
#  spectral/images in terms of VisAD data objects, and dataset query.
#  Note: MultiSpectralData is generic and can be subclassed for 
#  issues related to specific instrument data.
#- channelPropertyFile applies mainly to AIRS, but other instruments
#  may have the spectral files defined outside the primary data file.

def load(filename, channelPropertyFile=None):
  import os
  from os import listdir, path
  files = listdir(hydra_path+"/multispectral/properties") #- translation tables
  tablesList = []
  for x in files:
    if ( path.isfile(os.path.join(hydra_path+'/multispectral/properties',x)) == 1 ):
      table = AbstractNameAssociationTable(x)
      tablesList.append(table)

  import file
  from file import *

  readers     = file.__all__
  reader      = None
  readerTable = None

  for x in readers:
    try:
      reader = eval(x+"."+x+"(filename)")
    except:
      pass
    else:
      reader_name = x
      for y in tablesList:
        if y.verifyTable(reader) == true:
          readerTable = y
  
  if reader == None:
    raise Exception, "No local/remote read support for: "+filename

  if readerTable == None:
    print "No translation table found! Using default"
    readerTable = tablesList[files.index("default.txt")]


  instrName = readerTable.instrumentName

  if readerTable.instrumentNameAttribute != None:
    instrName = None
    try:
      instrName = reader.getGlobalAttr(readerTable.instrumentNameAttribute)
    except:
      instrName = reader.getValue(readerTable.instrumentNameAttribute, [0])
    else:
      pass
    
  #- check for an alias
  name = readerTable.p.getProperty(str(instrName))
  if name != None:
    instrName = name
    readerTable.instrumentName = instrName
  
    
  #-- reader needs access to name association table
  reader.readerTable = readerTable
  
  #-- Use adapter class. This forces specific reader to implement
  #   methods required by MultiSpectralData which will be
  #   referenced by applications.
  from file.MultiSpectralFile import MultiSpectralFile
  reader = MultiSpectralFile(reader)

  from ModisDataSource500m import ModisData500m
  from ModisDataSource250m import ModisData250m
  from ModisFireMask import ModisFireMaskData
  from SeaDASDataSource import SeaDASL2Data
 
  #-- most instruments can use the default, MultiSpectralData, but MODIS
  #   needs its own.
  if instrName   == "MODIS":
    return ModisData(filename, reader, readerTable)                     #- extends MultiSpectralData
  elif instrName == "MODIS_500m":
    return ModisData500m(filename, reader, readerTable) #- extends ModisData
  elif instrName == "MODIS_250m":
    return ModisData250m(filename, reader, readerTable) #- extends ModisData
  elif instrName == "AIRS":
    return AirsData(filename, reader, readerTable, channelPropertyFile) #- extends MultiSpectralData
  elif instrName == "MAS":
    return MasData(filename, reader, readerTable)
  elif instrName == "MSG_local":
    return MsgData(filename, reader, readerTable)
  elif instrName   == "MOD35":
    return ModisCloudMaskData(filename, reader, readerTable)
  elif instrName == "MOD14":
    return ModisFireMaskData(filename, reader, readerTable)
  elif instrName   == "MOD06":
    return Modis06Data(filename, reader, readerTable)
  elif instrName   == "MOD04":
    return Modis04Data(filename, reader, readerTable)
  elif instrName == "AIRSLevel2" or instrName == "AIRSUWLevel2" or instrName == "MODISUWLevel2" or instrName == "MODISLevel2":
    return Level2Data(filename, reader, readerTable)
  elif instrName == "WFC":
    return WFC_calipso(filename, reader, readerTable)
  elif instrName == "AVHRR_seaspace":
    return AVHRR_seaspace(filename, reader, readerTable)
  elif instrName == "SeaDAS_L2PROD1":
    return SeaDASL2Data(filename, reader, readerTable)
  elif instrName == "AMSU_A":
    return AMSU_A(filename, reader, readerTable)
  elif instrName == "GEOCAT":
    return GeoCat(filename, reader, readerTable)
  else:
    return MultiSpectralData(filename, reader, readerTable)
#---------------------------------------------------------------------------------------------

class MultiSpectralData:
  def __init__(self, filename, reader, readerTable, channelPropertyFile=None, master=None):
    if master == None:
      self.filename            = filename
      self.subset              = None
      self.reader              = reader
      self.geo_reader          = reader  #- geolocation may exist in separate file, eg. MOD03
      self.readerTable         = readerTable
      self.channelPropertyFile = channelPropertyFile
      self.instrumentName      = readerTable.instrumentName
      self.platformName        = None
      
      self.observationName     = (readerTable.observation).split(",")
      if readerTable.channelIndex != None:
        self.channelIndexName    = (readerTable.channelIndex).split(",")
      else:
        self.channelIndexName = None
      self.numObsDataSets      = len(self.observationName)
      self.numChannelsPerDataSet = []
      
      self.obsChannelsName     = readerTable.observationChannels
      if self.obsChannelsName != None:
        self.obsChannelsName = (self.obsChannelsName).split(",")
      self.obsChannelIntervalsName = readerTable.channelIntervals
      if self.obsChannelIntervalsName != None:
        self.obsChannelIntervalsName = (self.obsChannelIntervalsName).split(",")
        
      self.obsElementName      = (readerTable.obsElement).split(",")
      self.obsLineName         = (readerTable.obsLine).split(",")

      if readerTable.obsLevel != None:
         self.obsLevelName     = (readerTable.obsLevel).split(",")
      else:
         self.obsLevelName     = None

      self.geoLonElementName   = readerTable.geoLonElement
      self.geoLatElementName   = readerTable.geoLatElement
      self.geoLonLineName      = readerTable.geoLonLine
      self.geoLatLineName      = readerTable.geoLatLine
      self.obsLatitudeName     = readerTable.obsLatitude
      self.obsLongitudeName    = readerTable.obsLongitude
      self.segmentBounds       = readerTable.segmentBounds
      self.obsTimeOfDayName    = readerTable.obsTimeOfDay
      self.obsFOVangleName     = readerTable.obsFOVangle
      self.solarZenithName     = readerTable.solarZenith

      self.rangeUnitName = None
      if readerTable.range_units != None:
        self.rangeUnitName  = (readerTable.range_units).split(",")

      self.rangeLongName = None
      if readerTable.range_long_name != None:
        self.rangeLongName  = (readerTable.range_long_name).split(",")

      self.rangeValidMin = None
      if readerTable.range_valid_min != None:
        self.rangeValidMin  = (readerTable.range_valid_min).split(",")

      self.rangeValidMax = None
      if readerTable.range_valid_max != None:
        self.rangeValidMax  = (readerTable.range_valid_max).split(",")

      self.rangeMissing = None
      if readerTable.range_missing != None:
        self.rangeMissing = (readerTable.range_missing).split(",")

      self.rangeValidType = None
      if readerTable.range_valid_type != None:
        self.rangeValidType = (readerTable.range_valid_type).split(",")

      self.defaultObsUnit      = Parser.parse("mW/steradian/m^2/cm^-1")
      self.observationUnit     = []
      self.getObservationUnit()

      self.makeRealTypes()
      self.makeSpectrumDomainType()

      self.numObsElements      = self.reader.getNumObsElements(self.obsElementName[0])
      self.numObsLines         = self.reader.getNumObsLines(self.obsLineName[0])
      
      if self.channelIndexName != None:
        self.numObsChannels = 0
        for i in xrange(self.numObsDataSets):
          self.numChannelsPerDataSet.append(self.reader.getNumObsChannels(self.channelIndexName[i]))
          self.numObsChannels += self.numChannelsPerDataSet[i]
        self.obsDimList = [self.channelIndexName[0],self.obsElementName[0],self.obsLineName[0]]
      else:
        self.numObsChannels = 0
        for i in xrange(self.numObsDataSets):
          self.numObsChannels += 1 # assume a 2-D field with an extra singleton dimension
          self.numChannelsPerDataSet.append(1)
        self.obsDimList = [self.obsElementName[0],self.obsLineName[0]]

      if self.geoLonElementName != None and self.geoLonLineName != None:
        self.numGeoElements    = self.reader.getNumGeoElements(self.geoLonElementName)
        self.numGeoLines       = self.reader.getNumGeoLines(self.geoLonLineName)
      else:
        self.numGeoElements    = self.numObsElements
        self.numGeoLines       = self.numObsLines
        
      if self.numObsChannels <= 75:
        self.broadBand = true
      else:
        self.broadBand = false

      #-- for MODIS internal L1B low-res nav
      self.ratio  = self.numObsLines/self.numGeoLines

      self.makeTrackSegments()
  
    
      #- init geo range
      self.obsLonMinMax        = None
      self.obsLatMinMax        = None
      self.currentSegmentIndex = 0
      self.currentSpectrum     = None
      self.currentImage        = None
      self.currentImageCS      = None
      self.observationField    = None
      self.observationDomain   = None
    
      #----------------------------------------------------------------------------
      #-default(assumption)  0: channel, 1: element, 2: line;  rightmost varies fastest
      self.obsDimOrder         = self.reader.getDimOrder(self.observationName[0],self.obsDimList)
      self.xlength = self.numObsLines
      self.ylength = self.numObsElements
      self.spatialTupleArray = [self.obsLineType, self.obsElementType]
      self.invert = false
      jj = 0
      kk = 0
      if len(self.obsDimOrder) == 3:
        jj = 1
        kk = 2
      if len(self.obsDimOrder) == 2:
        jj = 0
        kk = 1
      if (self.obsDimOrder[jj] > self.obsDimOrder[kk]):
        self.xlength = self.numObsElements
        self.ylength = self.numObsLines
        self.spatialTupleArray[0] = self.obsElementType
        self.spatialTupleArray[1] = self.obsLineType
        self.invert = true

      if self.geoLatElementName != None and self.geoLatLineName != None:
        dim_list = [self.geoLatElementName, self.geoLatLineName]
      else:
        dim_list = [self.obsElementName[0], self.obsLineName[0]]
      self.geoDimOrder = self.reader.getDimOrder(self.obsLatitudeName, dim_list)
      #--------------------------------------------------------------------------

      #--- get channels arrays
      self.band_numbers                = None
      self.observationChannels         = None
      self.observationChannels         = self.getObservationChannels()
      self.observationChannelIntervals = None
      self.observationChannelIntervals = self.getObservationChannelIntervals()

      self.native_param = self.getNativeParam()
      self.paramName    = self.setDefaultParam()
      
      self.makeRangeTypes()
      self.makeSpectrumRangeType(self.paramName)
      self.makeImageRangeType(self.paramName)
      self.geoOffset = 0
      self.date_str  = None
      self.time_str  = None
      self.unnavigated = false
    else:
      self.copy(master)
  #-----------------------------------------------------------------------------
  
  def copy(self, master):
    self.filename            = master.filename
    self.reader              = master.reader
    self.geo_reader          = master.geo_reader
    self.readerTable         = master.readerTable
    self.channelPropertyFile = master.channelPropertyFile

    self.instrumentName      = master.instrumentName
    self.platformName        = master.platformName
    self.observationName     = master.observationName
    self.obsChannelsName     = master.obsChannelsName
    self.channelIndexName    = master.channelIndexName
    self.obsElementName      = master.obsElementName
    self.obsLineName         = master.obsLineName
    self.geoLonElementName   = master.geoLonElementName
    self.geoLatElementName   = master.geoLatElementName
    self.geoLonLineName      = master.geoLonLineName
    self.geoLatLineName      = master.geoLatLineName
    self.obsLatitudeName     = master.obsLatitudeName
    self.obsLongitudeName    = master.obsLongitudeName
    self.segmentBounds       = master.segmentBounds
    self.obsTimeOfDayName    = master.obsTimeOfDayName
    self.obsFOVangleName     = master.obsFOVangleName
    self.solarZenithName     = master.solarZenithName
    self.obsChannelIntervalsName = master.obsChannelIntervalsName
    self.rangeLongName     = master.rangeLongName
    self.rangeUnitName     = master.rangeUnitName
    self.rangeValidMin     = master.rangeValidMin
    self.rangeValidMax     = master.rangeValidMax
    self.rangeMissing      = master.rangeMissing
    self.rangeValidType    = master.rangeValidType


    self.observationType     = master.observationType
    self.obsChannelsIndexType = master.obsChannelsIndexType
    self.obsElementType      = master.obsElementType
    self.obsLineType         = master.obsLineType
    self.obsLongitudeType    = master.obsLongitudeType
    self.obsLatitudeType     = master.obsLatitudeType
    self.obsChannelsType     = master.obsChannelsType

    self.numObsElements      = master.numObsElements
    self.numObsLines         = master.numObsLines
    self.numObsChannels      = master.numObsChannels
    self.numGeoElements      = master.numGeoElements
    self.numGeoLines         = master.numGeoLines
    self.ratio               = master.ratio

    self.trackSegments       = master.trackSegments
    self.numberSegments      = master.numberSegments

    self.observationChannels = master.observationChannels
    self.observationChannelIntervals = master.observationChannelIntervals

    self.obsLonMinMax        = master.obsLonMinMax
    self.obsLatMinMax        = master.obsLatMinMax
    self.currentSegmentIndex = master.currentSegmentIndex
    self.currentSpectrum     = master.currentSpectrum
    self.currentImage        = master.currentImage
    self.currentImageCS      = master.currentImageCS
    self.observationField    = None
    self.obsDimOrder         = master.obsDimOrder
    self.xlength             = master.xlength
    self.ylength             = master.ylength
    self.invert              = master.invert
    self.spatialTupleArray   = master.spatialTupleArray
    self.geoDimOrder         = master.geoDimOrder
    self.fileIndexes         = master.fileIndexes
    self.native_param        = master.native_param
    self.paramName           = master.paramName
    self.numObsDataSets      = master.numObsDataSets
    self.numChannelsPerDataSet = master.numChannelsPerDataSet
    self.imageRangeType        = master.imageRangeType
    self.spectrumRangeType     = master.spectrumRangeType
    self.broadBand             = master.broadBand
    self.observationDomain     = master.observationDomain
    self.geoOffset             = master.geoOffset
    self.obsDimList            = master.obsDimList
    self.spectrumRangeTypeDict = master.spectrumRangeTypeDict
    self.imageRangeTypeDict    = master.imageRangeTypeDict
    self.date_str              = master.date_str
    self.time_str              = master.time_str
    self.defaultObsUnit        = master.defaultObsUnit
    self.observationUnit       = master.observationUnit
    self.unnavigated           = master.unnavigated
    
  def clone(self):
    return MultiSpectralData(self.filename,self.reader,self.readerTable,master=self)   
  
  def getNativeParam(self):
    native_param="radiance"
    return  native_param

  def setDefaultParam(self):
    return "brightnessTemp"

  def getObservationUnit(self):
    for i in xrange(self.numObsDataSets):
      unitStr = self.reader.getObsUnit(self.observationName[i], 'units')
      if unitStr == None:
        unit = self.defaultObsUnit
      else:
        try:
          unit = Parser.parse(unitStr)
        except:
          unit = self.defaultObsUnit
      self.observationUnit.append(unit)

  #--- VisAD RealTypes
  def makeRealTypes(self):
    name = self.readerTable.p.getProperty(self.observationName[0])
    if name == None:
      name = self.observationName[0]
    self.observationType      = getRealType(name)
    
    if self.channelIndexName != None:
      name = self.readerTable.p.getProperty(self.channelIndexName[0])
      if name == None:
        name = self.channelIndexName[0]
      self.obsChannelsIndexType = getRealType(name)
    else:
      self.obsChannelsIndexType = None
    
    name = self.readerTable.p.getProperty(self.obsElementName[0])
    if name == None:
      name = self.obsElementName[0]
    self.obsElementType       = getRealType(name)
    if self.obsElementType == None:
      self.obsElementType       = getRealType("xtrack")

    name = self.readerTable.p.getProperty(self.obsLineName[0])
    if name == None:
      name = self.obsLineName[0]
    self.obsLineType          = getRealType(name)
    if self.obsLineType == None:
      self.obsLineType = getRealType("track")
    
    self.obsLongitudeType     = RealType.Longitude
    self.obsLatitudeType      = RealType.Latitude
      
  def makeSpectrumDomainType(self):
    if self.obsChannelsName != None:
      centimeter = ScaledUnit(0.01, CommonUnit.meter, "cm")
      inv_cm = centimeter.pow(-1)
      inv_cm = ScaledUnit(1, inv_cm, "cm^-1")
      self.obsChannelsType = getRealType("wavenumber", inv_cm)
    elif self.obsChannelsIndexType != None:
      self.obsChannelsType = self.obsChannelsIndexType
    else:
      self.obsChannelsType = None

  def makeRangeTypes(self):
    self.rangeTypes = []
    self.rangeTypes.append(RealType.getRealType("radiance"))
    self.rangeTypes.append(RealType.getRealType("brightnessTemp"))    
    self.spectrumRangeTypeDict = {"radiance" : RealType.getRealType("radiance"), "brightnessTemp" : RealType.getRealType("brightnessTemp")}
    self.imageRangeTypeDict    = self.spectrumRangeTypeDict

  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = self.spectrumRangeTypeDict[paramName]
    
  def makeImageRangeType(self, paramName):
    self.imageRangeType = self.imageRangeTypeDict[paramName]
    
  def makeTrackSegments(self):
    if self.segmentBounds == None:
      self.trackSegments = None
    else:
      self.trackSegments = self.reader.getTrackSegments(self.segmentBounds)
    if (self.trackSegments != None):
      """
      any_bad = false
      for i in xrange(len(self.trackSegments)):
        if self.trackSegments[i] > self.numObsLines:
          any_bad = true
      if any_bad == true:
        self.numberSegments = 1
        self.trackSegments = [1, self.numObsLines]
      else:
      """
      self.numberSegments = len(self.trackSegments)/2
    else: #-- not defined, so set to 1
      self.numberSegments = 1
      #self.trackSegments = [0, self.numObsLines-1]
      #- one's based here
      self.trackSegments = [1, self.numObsLines]
    
  def getObservationChannels(self):
    self.fileIndexes = None
    if self.observationChannels == None:
      if self.obsChannelsName != None:
        self.observationChannels = []
        for i in xrange(self.numObsDataSets):
          self.observationChannels += list(self.reader.getValues(self.obsChannelsName[i]))
      else:
        self.observationChannels = [i for i in xrange(self.numObsChannels)]
    
    return self.observationChannels

  def getObservationChannelIntervals(self): 
    if self.observationChannelIntervals == None:
      if self.obsChannelIntervalsName != None:
        self.observationChannelIntervals = self.reader.getValues(self.obsChannelIntervalsName)
    return self.observationChannelIntervals

  def getObservations(self, line, element, copy=false):

    missing = false
    if (line < 0 or line > self.xlength-1) or (element < 0 or element > self.ylength-1):
      missing = true

    if (self.invert == true):
      tmp     = element
      element = line
      line    = tmp

    self.imageLine = line
    self.imageElement = element

    line, element = self.missingObs.dataToSegment(line,element)
      
    if (self.subset != None):      
      line    *= self.step
      element *= self.step
      if (self.invert == true):
        line     += self.subset.start_y
        element  += self.subset.start_x
      else:
        line     += self.subset.start_x
        element  += self.subset.start_y
    else:
      line    *= self.step
      element *= self.step
      
    ll = int(line + (self.trackSegments[self.currentSegmentIndex*2]-1)) #convert one to zero based
   
    #-- file access  
    vals = jarray.zeros(self.numObsChannels, 'f')
    cnt = 0
    for i in xrange(self.numObsDataSets):
      if len(self.obsDimList) == 3:
        start  = [0,element+self.geoOffset,ll+self.geoOffset]
        count  = [self.numChannelsPerDataSet[i],1,1]
        nstart = [None, None, None]
        ncount = [None, None, None]
      elif len(self.obsDimList) == 2:
        start  = [element+self.geoOffset,ll+self.geoOffset]
        count  = [1,1]
        nstart = [None, None]
        ncount = [None, None]
      for j in xrange(len(start)):
        nstart[self.obsDimOrder[j]]  = start[j]
        ncount[self.obsDimOrder[j]]  = count[j]
        
      if missing == true:
        vv = jarray.zeros(self.numChannelsPerDataSet[i], 'f')
        for ii in xrange(len(vv)):
          vv[ii] = Float.NaN
      else:
        vv = self.reader.getValues(self.observationName[i], nstart, ncount)      
      for k in xrange(len(vv)):
        vals[cnt] = float(vv[k])
        cnt += 1
        
    vals = self.getSpectrumRange(vals, self.paramName)
    
    #-- make the VisAD data object
    if self.observationField == None or copy == true: 
      self.observationDomain = Gridded1DSet(self.obsChannelsType, [self.observationChannels], self.numObsChannels)
      
      if self.broadBand == false:
        obs = FlatField(FunctionType(self.obsChannelsType, self.spectrumRangeType), self.observationDomain)
        obs.setSamples([vals])
        if copy == false:
          self.observationField = obs
      else:
        if (isinstance(self.spectrumRangeType, RealTupleType)):
          fields = []
          for i in xrange(self.spectrumRangeType.getDimension()):
            range_type = RealTupleType([self.obsChannelsType]+[self.spectrumRangeType.getComponent(i)])
            ff = FlatField(FunctionType(RealType.Generic, range_type), Integer1DSet(self.numObsChannels))
            ff.setSamples([self.observationChannels, vals[i]])
            fields.append(ff)
          obs = Tuple(fields)
          if copy == false:
            self.observationField = obs
        else:
          range_type = RealTupleType(self.obsChannelsType, self.spectrumRangeType)
          obs = FlatField(FunctionType(RealType.Generic, range_type ), Integer1DSet(self.numObsChannels))
          obs.setSamples([self.observationChannels, vals])
          if copy == false:
            self.observationField = obs
      return obs
    else:
      if self.broadBand == false:
        self.observationField.setSamples([vals])
      else:
        if (isinstance(self.spectrumRangeType, RealTupleType)):
          for i in xrange(self.spectrumRangeType.getDimension()):
            self.observationField.getComponent(i).setSamples([self.observationChannels, vals[i]])
        else:
          self.observationField.setSamples([self.observationChannels, vals])
    
    return self.observationField
  
  # default of this method returns the data as is, changes for Level 2 Products
  def extractValues(self,dataToExtract):
    return dataToExtract
    
  def getImageSegment(self, channelIndex, segmentIndex=0, subset=None):
    if channelIndex < 0:
      return None
    dataset_index = self.channelIndexToDataSetIndex(channelIndex)
    channelIndex  = self.channelIndexToFileIndex(dataset_index, channelIndex)
    self.currentSegmentIndex = segmentIndex   

    segStartLine = int(self.trackSegments[self.currentSegmentIndex*2]-1)  #convert one to zero based
    segEndLine   = int(self.trackSegments[self.currentSegmentIndex*2+1]-1)
    
    if subset != None: #-- subset logic
      self.subset = subset      
      if subset.end_x == None or subset.end_y == None:
        self.startElem = 0
        self.nElements = int((self.numObsElements)/subset.step)
        self.startLine = segStartLine
        self.nLines    = int((int(segEndLine-self.startLine+1))/subset.step)
      else:                 
        t_line  = subset.start_x
        t_elem  = subset.start_y
        n_lines = int((subset.end_x - subset.start_x)/subset.step) + 1
        n_elems = int((subset.end_y - subset.start_y)/subset.step) + 1
        if (self.invert == true):
          t_elem  = subset.start_x
          t_line  = subset.start_y
          n_elems = int((subset.end_x - subset.start_x)/subset.step) + 1
          n_lines = int((subset.end_y - subset.start_y)/subset.step) + 1
        
        self.startElem = t_elem
        self.startLine = segStartLine + t_line      
        self.nElements = n_elems
        self.nLines    = n_lines
      
      self.step      = self.subset.step
           
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    else:
      self.subset    = None
      self.startElem = 0
      self.nElements = self.numObsElements
      self.startLine = segStartLine
      self.nLines    = int(segEndLine-self.startLine+1)
      
      self.step      = 1  
      
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    if (len(self.obsDimList) == 3):
      start   = [channelIndex,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
      count   = [1,self.nElements,self.nLines]
      stride  = [1, self.step, self.step]
      nstart  = [None, None, None]
      ncount  = [None, None, None]
      nstride = [None, None, None]
    elif len(self.obsDimList) == 2:
      start   = [self.startElem+self.geoOffset,self.startLine+self.geoOffset]
      count   = [self.nElements,self.nLines]
      stride  = [self.step, self.step]
      nstart  = [None, None]
      ncount  = [None, None]
      nstride = [None, None]

    for i in xrange(len(start)):
      nstart[self.obsDimOrder[i]]  = start[i]
      ncount[self.obsDimOrder[i]]  = count[i]
      nstride[self.obsDimOrder[i]] = stride[i]

    if self.step > 1: #- this hack will go away when Java-Netcdf implements stride reading
      self.z = self.reader.getValues(self.observationName[dataset_index], nstart, ncount, nstride)
    else:
      self.z = self.reader.getValues(self.observationName[dataset_index], nstart, ncount, None)
    
#   call used to extract data from level 2 data currently
    self.z=self.extractValues(self.z)
        
    # Read in geographic data
    start  =  [self.startElem/self.ratio, self.startLine/self.ratio]
    count  =  [nGeoElements, nGeoLines]
    stride =  [self.geo_step, self.geo_step]
    nstart =  []
    ncount =  []
    nstride = []
    
    for i in xrange(len(start)):
      nstart.append(start[self.geoDimOrder[i]])
      ncount.append(count[self.geoDimOrder[i]])
      nstride.append(stride[self.geoDimOrder[i]])
      
    if self.step > 1:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, nstride)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, nstride)
    else:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, None)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, None)


    if self.solarZenithName != None:
      if self.step > 1:
        self.solarZenith = self.getSolarZenith(self.solarZenithName, nstart, ncount, nstride)
      else:
        self.solarZenith = self.getSolarZenith(self.solarZenithName, nstart, ncount, None)


    if self.invert == true:
      self.ylength  = self.nLines
      self.xlength  = self.nElements
      self.gylength = nGeoLines
      self.gxlength = nGeoElements
    else:
      self.xlength  = self.nLines
      self.ylength  = self.nElements
      self.gxlength = nGeoLines
      self.gylength = nGeoElements

    self.missingObs = self.initMissing(self.lons, self.lats, self.z)
    self.xlength, self.ylength = self.missingObs.getLengths()    
    self.z    = self.missingObs.removeMissing(self.z)
    self.lons = self.missingObs.removeMissing(self.lons)
    self.lats = self.missingObs.removeMissing(self.lats)
    
    if self.ratio == 1:
      self.gxlength = self.xlength
      self.gylength = self.ylength
     
    try:
      self.gs = Gridded2DSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
              [self.lats, self.lons],
              self.gxlength,self.gylength,None,None,None,0,0)
      self.cs = GridCoordinateSystem(self.gs)
    except:
      try:
         self.gs = Gridded2DDoubleSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
                                      [self.lats, self.lons],
                                      self.gxlength,self.gylength,None,None,None,0)
      except Exception, e:
        print e
        self.cs = None
        print "*** Error Making Image CoordinateSystem ***"
        
    self.cs = GridCoordinateSystem(self.gs)

    if self.unnavigated == false:
      self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs,None)
    else:
      self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1])

    self.imageDomain  = makeDomain(self.imageDomType,
          0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
          0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      
    self.subsetDomain = makeDomain(self.imageDomType,
          0, (self.xlength-1)*self.step, self.xlength,
          0, (self.ylength-1)*self.step, self.ylength)
          
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    self.imageField        = FlatField(self.imageFunctionType, self.imageDomain)  
    self.z                 = self.getImageRange(self.z,self.paramName,dataset_index,channelIndex)
        
    indexes                = jarray.zeros(2, 'i')
    minmax                 = Helper.minmax(self.z, indexes)
    minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
    maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
    self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
    self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
    self.imageField.setSamples([self.z], false)
       
    return self.imageField

  def setUnnavigatedImage(self, flag):
    self.unnavigated = flag

  def getUnnavigatedImage(self):
    self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1])
    self.imageDomain  = makeDomain(self.imageDomType,
          0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
          0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    self.imageField = FlatField(self.imageFunctionType, self.imageDomain)
    self.imageField.setSamples([self.z], false)
    return self.imageField

  def getNavigatedImage(self, replace=true):
    if replace == true:
      self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs, None)
      self.imageDomain  = makeDomain(self.imageDomType,
            0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
            0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
      self.imageField = FlatField(self.imageFunctionType, self.imageDomain)
      self.imageField.setSamples([self.z], false)
      return self.imageField
    else:
      imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs, None)
      imageDomain  = makeDomain(imageDomType,
            0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
            0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      imageFunctionType = FunctionType(imageDomType, self.imageRangeType)
      imageField = FlatField(imageFunctionType, imageDomain)
      imageField.setSamples([self.z], false)
      return imageField

  def getSameImageSegment(self, channelIndex):
    if channelIndex >= 0:
      dataset_index = self.channelIndexToDataSetIndex(channelIndex)
      channelIndex  = self.channelIndexToFileIndex(dataset_index, channelIndex)

      if len(self.obsDimList) == 3:
        start  =  [channelIndex,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count  =  [1,self.nElements,self.nLines]
        stride =  [1, self.step, self.step]
        nstart  = [None, None, None]
        ncount  = [None, None, None]
        nstride = [None, None, None]
      elif len(self.obsDimList) == 2:
        start  =  [self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count  =  [self.nElements,self.nLines]
        stride =  [self.step, self.step]
        nstart  = [None, None]
        ncount  = [None, None]
        nstride = [None, None]

      for j in xrange(len(start)):
        nstart[self.obsDimOrder[j]]  = start[j]
        ncount[self.obsDimOrder[j]]  = count[j]
        nstride[self.obsDimOrder[j]] = stride[j]
       
      if self.step > 1: 
        self.z = self.reader.getValues(self.observationName[dataset_index],nstart,ncount,nstride)
      else:
        self.z = self.reader.getValues(self.observationName[dataset_index],nstart,ncount)
        
      self.z = self.missingObs.removeMissing(self.z)
      self.z = self.getImageRange(self.z, self.paramName, dataset_index, channelIndex)
      
      indexes                = jarray.zeros(2, 'i')
      minmax                 = Helper.minmax(self.z, indexes)
      minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
      maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
      self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
      self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
      
      self.imageField.setSamples([self.z], false)


  def getSameImageSegmentCopy(self, channelIndex):
    if channelIndex >= 0:
      dataset_index = self.channelIndexToDataSetIndex(channelIndex)
      channelIndex  = self.channelIndexToFileIndex(dataset_index, channelIndex)

      if len(self.obsDimList) == 3:
        start   =  [channelIndex,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count   =  [1,self.nElements,self.nLines]
        stride  =  [1,self.step,self.step]
        nstart  = [None, None, None]
        ncount  = [None, None, None]
        nstride = [None, None, None]
      elif len(self.obsDimList) == 2:
        start   =  [self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count   =  [self.nElements,self.nLines]
        stride  =  [self.step,self.step]
        nstart  = [None, None]
        ncount  = [None, None]
        nstride = [None, None]

      for j in xrange(len(start)):
        nstart[self.obsDimOrder[j]]  = start[j]
        ncount[self.obsDimOrder[j]]  = count[j]
        nstride[self.obsDimOrder[j]] = stride[j]
        
      if self.step > 1:  
        self.z = self.reader.getValues(self.observationName[dataset_index],nstart,ncount,nstride)
      else:
        self.z = self.reader.getValues(self.observationName[dataset_index],nstart,ncount)
        
      self.z = self.missingObs.removeMissing(self.z)
      new_imageField = FlatField(self.imageFunctionType, self.imageDomain)
      self.z = self.getImageRange(self.z, self.paramName, dataset_index, channelIndex)
      
      indexes                = jarray.zeros(2, 'i')
      minmax                 = Helper.minmax(self.z, indexes)
      minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
      maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
      self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
      self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
      
      new_imageField.setSamples([self.z], false)
      
      return new_imageField
  
  # this method is currently for MOD06 and MOD07 files
  def getLevel2ImageSegment(self,segmentIndex=0,subset=None):  
    if self.paramName == None:
      return None
    
    self.fieldToRetrieve=self.imageRangeTypeDict.get(self.paramName).toString()
    self.makeImageRangeType(self.paramName)
    self.currentSegmentIndex = segmentIndex   
        
    segStartLine = int(self.trackSegments[self.currentSegmentIndex*2]-1)  #convert one to zero based
    segEndLine   = int(self.trackSegments[self.currentSegmentIndex*2+1]-1)
    
    if subset != None: #-- subset logic
      self.subset = subset      
      if subset.end_x == None or subset.end_y == None:
        self.startElem = 0
        self.nElements = int((self.numObsElements)/subset.step)
        self.startLine = segStartLine
        self.nLines    = int((int(segEndLine-self.startLine+1))/subset.step)
      else:                 
        t_line  = subset.start_x
        t_elem  = subset.start_y
        n_lines = int((subset.end_x - subset.start_x)/subset.step) + 1
        n_elems = int((subset.end_y - subset.start_y)/subset.step) + 1
        if (self.invert == true):
          t_elem  = subset.start_x
          t_line  = subset.start_y
          n_elems = int((subset.end_x - subset.start_x)/subset.step) + 1
          n_lines = int((subset.end_y - subset.start_y)/subset.step) + 1
        
        self.startElem = t_elem
        self.startLine = segStartLine + t_line      
        self.nElements = n_elems
        self.nLines    = n_lines
      
      self.step      = self.subset.step
           
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    else: #  not a subset
      self.subset    = None
      self.startElem = 0
      self.nElements = self.numObsElements
      self.startLine = segStartLine
      self.nLines    = int(segEndLine-self.startLine+1)
      
      self.step      = 1  
      
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    start   = [self.startElem,self.startLine]
    count   = [self.nElements,self.nLines]
    stride  = [self.step, self.step]
      
    nstart  = [None, None]
    ncount  = [None, None]
    nstride = [None, None]

    for i in xrange(len(start)):
      nstart[self.obsDimOrder[i]]  = start[i]
      ncount[self.obsDimOrder[i]]  = count[i]
      nstride[self.obsDimOrder[i]] = stride[i]
    
    if self.step > 1: #- this hack will go away when Java-Netcdf implements stride reading
      self.z = self.reader.getValues(self.fieldToRetrieve, nstart, ncount, nstride)
    else:
      self.z = self.reader.getValues(self.fieldToRetrieve, nstart, ncount, None)
    
#   call used to extract data from level 2 data currently
    self.z=self.extractValues(self.z)
        
    # Read in geographic data  
    start  =  [self.startElem/self.ratio, self.startLine/self.ratio] 
    count  =  [nGeoElements, nGeoLines]
    stride =  [self.geo_step, self.geo_step]
    
    nstart =  []
    ncount =  []
    nstride = []
    
    for i in xrange(len(start)):
      nstart.append(start[self.geoDimOrder[i]])
      ncount.append(count[self.geoDimOrder[i]])
      nstride.append(stride[self.geoDimOrder[i]])
      
    if self.step > 1:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, nstride)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, nstride)
    else:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, None)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, None)
    
    if self.invert == true:
      self.ylength  = self.nLines
      self.xlength  = self.nElements
      self.gylength = nGeoLines
      self.gxlength = nGeoElements
    else:
      self.xlength  = self.nLines
      self.ylength  = self.nElements
      self.gxlength = nGeoLines
      self.gylength = nGeoElements

      
    self.missingObs = self.initMissing(self.lons, self.lats, self.z)
    self.xlength, self.ylength = self.missingObs.getLengths()    
    self.z    = self.missingObs.removeMissing(self.z)
    self.lons = self.missingObs.removeMissing(self.lons)
    self.lats = self.missingObs.removeMissing(self.lats)
    
    if self.ratio == 1:
      self.gxlength = self.xlength
      self.gylength = self.ylength
     
    try:
      self.gs = Gridded2DSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
              [self.lats, self.lons],
              self.gxlength,self.gylength,None,None,None,0,0)
      self.cs = GridCoordinateSystem(self.gs)
    except:
      self.cs = None
      print "*** Error Making Image CoordinateSystem ***"
        
    self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs,None)
    self.imageDomain  = makeDomain(self.imageDomType,
          0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
          0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      
      
    self.subsetDomain = makeDomain(self.imageDomType,
          0, (self.xlength-1)*self.step, self.xlength,
          0, (self.ylength-1)*self.step, self.ylength)
          
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    self.imageField        = FlatField(self.imageFunctionType, self.imageDomain)  
    self.z                 = self.getImageRange(self.z)
    
    indexes                = jarray.zeros(2, 'i')
    minmax                 = Helper.minmax(self.z, indexes)
    minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
    maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
    self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
    self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
    self.imageField.setSamples([self.z], false)
       
    return self.imageField    
      
  def changeParameter(self, paramName, channelIndex, line, element):
    self.paramName         = paramName
    self.spectrumRangeType = self.spectrumRangeTypeDict[paramName]
    self.imageRangeType    = self.imageRangeTypeDict[paramName]
        
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    
    self.imageField        = self.getSameImageSegmentCopy(channelIndex)
    self.observationField  = self.getObservations(line, element, true)
        
    return self.imageField, self.observationField
     
  def validateEarthPointOnImage(self, point):
    if point == None:
      return false
    comps = point.getComponents()
    lon = comps[0].getValue()
    lat = comps[1].getValue()
    if (Double.isNaN(lon) == 1 or Double.isNaN(lat) == 1):
      return false
    xy = self.cs.fromReference([[lat], [lon]])
    if ((xy[0][0] >= 0 and xy[0][0] <= self.gxlength) and (xy[1][0] >= 0 and xy[1][0] <= self.gylength)):
      return true
    else:
      return false

  def getGeoDomain(self):
    if self.obsLonMinMax == None:      
      start  = [0,0]
      count  = [self.numGeoElements, self.numGeoLines]
      nstart = []
      ncount = []
      for i in xrange(len(start)):
        nstart.append(start[self.geoDimOrder[i]])
        ncount.append(count[self.geoDimOrder[i]])
      
      # do a subset of the image
      if self.numGeoElements > 512:
        ncount[0] = ncount[0]/10
        ncount[1] = ncount[1]/10
        nstride = [10,10]
      else:
        nstride = None
        
      lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, nstride)
      lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, nstride)
      
      self.obsLonMinMax = Helper.minmax(lons, -180, 180)
      self.obsLatMinMax = Helper.minmax(lats, -90, 90)
    return self.obsLatMinMax, self.obsLonMinMax 
    
      
  def getTimeObs(self, line, element):
    if (self.invert == true):
      tmp     = element
      element = line
      line    = tmp
    line, element = self.missingObs.dataToSegment(line, element)
    
    if (self.subset != None):      
      if (self.invert == true):
        line     += self.subset.start_y
        element  += self.subset.start_x
      else:
        line     += self.subset.start_x
        element  += self.subset.start_y
        
    ll = int(line + (self.trackSegments[self.currentSegmentIndex*2]-1))
    
    return self.reader.getValue(self.obsTimeOfDayName, [element, ll])
    

  def getFOVangleObs(self, line, element):
    if (self.invert == true):
      tmp     = element
      element = line
      line    = tmp
    line, element = self.missingObs.dataToSegment(line,element)
    
    if (self.subset != None):      
      if (self.invert == true):
        line     += self.subset.start_y
        element  += self.subset.start_x
      else:
        line     += self.subset.start_x
        element  += self.subset.start_y
        
    ll = int(line + (self.trackSegments[self.currentSegmentIndex*2]-1))    
    return self.reader.getValue(self.obsFOVangleName, [element, ll])
      
  def getImageValue(self, lat, lon):
    if self.unnavigated == false:
      val = self.imageField.evaluate(RealTuple(self.cs.getReference(), [lat, lon]), Data.NEAREST_NEIGHBOR, Data.NO_ERRORS)
    else:
      cc = self.cs.fromReference([[lat],[lon]])
      val = self.imageField.evaluate(RealTuple(self.imageDomType, [cc[0][0], cc[1][0]]), Data.NEAREST_NEIGHBOR, Data.NO_ERRORS)
    return val
  
  def getScaleOffset(self):
    self.paramScale   = []
    self.paramOffset  = []
        
    for i in xrange(self.numObsDataSets):
      paramScale, paramOffset   = self.reader.getObsScaleOffset(self.observationName[i], "scale_factor", "add_offset")
      if paramScale == None or paramOffset == None:
        self.paramScale.append(None)
        self.paramOffset.append(None)
      else:
        scales  = jarray.zeros(len(paramScale), 'f')
        offsets = jarray.zeros(len(paramOffset), 'f')
        for i in xrange(len(scales)):
          scales[i]  = paramScale[i]
          offsets[i] = paramOffset[i]
        
        self.paramScale.append(scales)
        self.paramOffset.append(offsets)
        
  def getRealTypes(self):
    return [self.obsElementType, self.obsLineType, None, self.obsChannelsType, self.imageRangeType]

  def getGeoRealTypes(self):
    return [self.obsLatitudeType, self.obsLongitudeType]

  def getObservationChannelRange(self):
    if self.broadBand == false:
      return (min(self.observationChannels), max(self.observationChannels))
    else: #- return band numbers, ie MODIS or GOES
      return (min(self.observationChannels), max(self.observationChannels))
    
  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
    if self.paramName   == "radiance":
      if self.observationUnit[0].equals(self.defaultObsUnit):
        return [0, 150]
      elif self.observationUnit[0].equals(Parser.parse("W/steradian/m^2/cm^-1")):
        return [0, 0.150]
      elif self.observationUnit[0].equals(Parser.parse("W/m^2/steradian/micrometer")):
        return [0, 1200]
    elif self.paramName == "brightnessTemp":
      if self.observationUnit[0].equals(Parser.parse("K")):
        return [180, 320]
      elif self.observationUnit[0].equals(Parser.parse("C")):
        return [-90, 50]
      else:
        return [180, 320]
      
  def getNumSegments(self):
    return self.numberSegments
    
  def initMissing(self, longitudes, latitudes, values):
    class Missing: #- this class must implement these methods
      def __init__(self, longitudes, latitudes, values, xlength, ylength, invert):
        self.longitudes = longitudes
        self.latitudes  = latitudes
        self.xlength    = xlength
        self.ylength    = ylength
        self.invert     = invert
        
      def dataToSegment(self, line, element):
        return line, element
        
      def removeMissing(self, values):
        return values
        
      def getLengths(self):
        return self.xlength, self.ylength
      
    m = Missing(longitudes, latitudes, values, self.xlength, self.ylength, self.invert)
    return m
    
  def channelIndexToFileIndex(self, dataset_index, channel_index):
    sum = 0
    for i in xrange(dataset_index):
      sum += self.numChannelsPerDataSet[i]
    return channel_index - sum
    
  def channelIndexToDataSetIndex(self, index):
    start = 0
    for i in xrange(self.numObsDataSets):
      if (index >= start and index <= start + self.numChannelsPerDataSet[i]-1):
        return i
      else:
        start += self.numChannelsPerDataSet[i]
    return -1
  
  def getSpectrumRange(self, values, paramName):
    if self.native_param == paramName:
      return values
    elif paramName == "brightnessTemp":
      return self.radianceToBrightnessTempSpectrum(values)
    elif paramName == "radiance":
      return self.brightnessTempToRadiance(values,channelIndex)
      
  def getImageRange(self, values, paramName, datasetIndex, channelIndex):
    if self.native_param == paramName:
      self.imageRangeUnit = self.observationUnit[0]
      return values
    elif paramName == "brightnessTemp":
      self.imageRangeUnit = Parser.parse("K")
      return self.radianceToBrightnessTemp(values,channelIndex)
    elif paramName == "radiance":
      self.imageRangeUnit = ScaledUnit(1, self.defaultObsUnit, "mW/ster/m^2/cm^-1")
      return self.brightnessTempToRadiance(values,channelIndex)
           
  def radianceToBrightnessTemp(self, values, channelIndex):
    # Converts radiances [mW/ster/m2/cm^-1] to BT [K]
    # Input: nu  array of wavenmbers [cm^-1]
    #        B   radiances [mW/ster/m2/cm^-1]
    # Output: bt brightness temperature in [K]
    # Paolo Antonelli
    # Wed Feb 25 16:43:05 CST 1998

    thisUnit = Parser.parse("mW/steradian/m^2/cm^-1")
    values = thisUnit.toThis(values, self.observationUnit[self.channelIndexToDataSetIndex(channelIndex)])
 
    c1=1.191066E-5;           #- mW/m2/ster/cm^-4
    c2=1.438833;              #- K*cm
    wavenumber = self.observationChannels[channelIndex]
    nu         = wavenumber
    nvals      = len(values)
    new_values = jarray.zeros(nvals, 'f')
    for i in xrange(nvals):
      B = values[i]
      K = (c1*nu*nu*nu)/B
      if K == 0.0:
        bt = B
      else:
        bt=c2*nu/(Math.log(((c1*nu*nu*nu)/B)+1.0));
      new_values[i] = bt
      
    return new_values
    
  def radianceToBrightnessTempSpectrum(self, values):
    # Converts radiances [mW/ster/m2/cm^-1] to BT [K]
    # Input: nu  array of wavenmbers [cm^-1]
    #        B   radiances [mW/ster/m2/cm^-1]
    # Output: bt brightness temperature in [K]
    # Paolo Antonelli
    # Wed Feb 25 16:43:05 CST 1998

    thisUnit = Parser.parse("mW/steradian/m^2/cm^-1")
    values = thisUnit.toThis(values, self.observationUnit[0])

    c1=1.191066E-5;           #- mW/m2/ster/cm^-4
    c2=1.438833;              #- K*cm
                                                
    nvals = len(values)
    new_values = jarray.zeros(nvals, 'f')   
    for i in xrange(nvals):
      nu = self.observationChannels[i]
      B = values[i]
      bt=c2*nu/(Math.log(((c1*nu*nu*nu)/B)+1.0));
      new_values[i] = bt
      
    return new_values
    
  def brightnessTempToRadiance(self, values, wavenumber):
    # compute radiance given wavenumbers and brightness temperature.
    #
    # Inputs:
    #	freq	  wavenumbers (Nx1) in 1/cm 
    #	bt        brightnes temperature (Nx1) in Kelvin
    # Outputs:
    #	radiance  Planck radiances (Nx1) in mW/(m2 sr. 1/cm)
    #
    # see also RADTOT.M, RAD2BT.M, TTORAD.M
    # 
    # DCT 11/11/99
    #

    # fundamental constants:
    #  (Cohen, E.R. and B.N. Taylor, The 1986 CODATA recommended values
    #  of the fundamental physical constants, Journal of Research of
    #  the National Bureau of Standard, 92(2), March-April 1987.)

    h = 6.6260755E-34;  # Planck constant in Js
    c = 2.99792458E8;   # photon speed in m/s
    k = 1.380658E-23;   # Boltzmann constant in J/K

    c1 = 2*h*c*c
    c2 = h*c/k

    freq  = wavenumber*1.0e2
    nvals = len(values)
    new_values = jarray.zeros(nvals, 'f')
    for i in xrange(nvals):
      bt = values[i]
      radiance = 1e5 * c1*(freq*freq*freq)/(Math.exp((c2*freq)/bt)-1);
      new_values[i] = float(radiance)

    return new_values

  def getInitialSubset(self):
    return None

  def getBoundingBox(self):
    return None

#------------------------------------------------------------------------------------------------   
#- eventually break these out into separate modules
class ModisData(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in same directory as L1B
    if master == None:
      if filename.find("1000m") != -1:
        geo_filename = filename.replace("1000m", "geo")
      elif filename.find("MOD02") != -1:
        geo_filename = filename.replace("MOD02", "MOD03")
      elif filename.find("MYD02") != -1:
        geo_filename = filename.replace("MYD02", "MYD03")
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"
    
      import file
      from file import *
      from java.lang import String
      
      self.getScaleOffset()
      self.geoOffset = 2

      try:
        self.date_str = reader.getGlobalAttr("RANGEBEGINNINGDATE")
        self.time_str = reader.getGlobalAttr("RANGEBEGINNINGTIME")
        self.date_str = str(self.date_str)
        self.time_str = str(self.time_str)
      except:
        self.date_str = None
        self.time_str = None

      if self.date_str == None or self.time_str == None:
        try:
          attr = reader.getGlobalAttr(2)
        except:
          attr = reader.getGlobalAttr(4)
        ss = str(attr)

        i = ss.find('RANGEBEGINNINGDATE')
        ii = ss.find('END_OBJECT', i)
        new_ss = attr.substring(i,ii)
        ssa=new_ss.splitlines()
        self.date_str = (((ssa[2].split('=')))[1].split('"'))[1]

        i = ss.find('RANGEBEGINNINGTIME')
        ii = ss.find('END_OBJECT', i)
        new_ss = attr.substring(i,ii)
        ssa=new_ss.splitlines()
        self.time_str = (((ssa[2].split('=')))[1].split('"'))[1]

    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
      self.radScale       = master.radScale
      self.radOffset      = master.radOffset
      self.reflScale      = master.reflScale
      self.reflOffset     = master.reflOffset
      self.geoOffset      = master.geoOffset
      self.date_str       = master.date_str
      self.time_str       = master.time_str
      
      
  def getObservationChannels(self):     
    channels = []
    self.band_numbers = []
    ff       = open(hydra_path+'/ancillary/modis/default_cntr_wn_bw.txt', 'r')
    lines    = ff.readlines()
    nlines   = len(lines)
    self.bandDict = {}
    cnt = 0
    for i in xrange(nlines):
      if lines[i].startswith('!'):
        pass
      else:
        fields      = lines[i].split()
        band_number = fields[0]
        center_wn   = fields[1]
        low_wn      = fields[2]
        hi_wn       = fields[3]
        center_wn   = float(center_wn)
        if band_number == "13lo":
          center_wn -= center_wn/1000
        if band_number == "13hi":
          center_wn += center_wn/1000
        if band_number == "14lo":
          center_wn -= center_wn/1000
        if band_number == "14hi":
          center_wn += center_wn/1000
        self.band_numbers.append(band_number)
        channels.append(center_wn)
        self.bandDict.update({band_number : [cnt,center_wn]})
        self.bandDict.update({center_wn : band_number})
        cnt += 1
        
    self.fileIndexes = QuickSort.sort(channels)
    new_channels     = jarray.zeros(len(channels), 'f')
    
    for i in xrange(len(new_channels)):
      new_channels[i] = channels[self.fileIndexes[i]]
    return new_channels
       
  #- override
  def makeTrackSegments(self):
    if self.numObsLines < 2600:
      self.numberSegments = 1
      self.trackSegments = [1, self.numObsLines]     
      return
    """  
    self.trackSegments = []
    segLen = 2000
    start  = 401  #-track segment array is one's based 
    self.numberSegments = int((self.numObsLines-100)/segLen)
    for i in xrange(self.numberSegments):
      start = start + segLen*i
      self.trackSegments.append(start)
      self.trackSegments.append(start + (segLen-1))
    """
    self.numberSegments = 1
    self.trackSegments = [401, self.numObsLines-100]

  def getInitialSubset(self):
    return Subset(None,None,None,None,step=10)
      
        
  def getScaleOffset(self):
    self.radScale   = []
    self.radOffset  = []
    self.reflScale  = []
    self.reflOffset = []
    
    for i in xrange(self.numObsDataSets):
      radScale, radOffset   = self.reader.getObsScaleOffset(self.observationName[i], "radiance_scales", "radiance_offsets")
      reflScale, reflOffset = self.reader.getObsScaleOffset(self.observationName[i], "reflectance_scales", "reflectance_offsets")
      if radScale == None or radOffset == None:
        self.radScale.append(None)
        self.radOffset.append(None)
      else:
        scales  = jarray.zeros(len(radScale), 'f')
        offsets = jarray.zeros(len(radOffset), 'f')
        for i in xrange(len(scales)):
          scales[i]  = radScale[i]
          offsets[i] = radOffset[i]
        self.radScale.append(scales)
        self.radOffset.append(offsets)
        
      if reflScale == None or reflOffset == None:
        self.reflScale.append(None)
        self.reflOffset.append(None)
      else:
        scales  = jarray.zeros(len(reflScale), 'f')
        offsets = jarray.zeros(len(reflOffset), 'f')
        for i in xrange(len(scales)):
          scales[i]  = reflScale[i]
          offsets[i] = reflOffset[i]
        self.reflScale.append(scales)
        self.reflOffset.append(offsets)
  
      
  def getSpectrumRange(self, values, paramName):
    if paramName == "radiance":
      rad_scales  = jarray.zeros(self.numObsChannels, 'f')
      rad_offsets = jarray.zeros(self.numObsChannels, 'f')
      cnt = 0
      for i in xrange(self.numObsDataSets):
        System.arraycopy(self.radScale[i], 0, rad_scales, cnt, self.numChannelsPerDataSet[i])
        System.arraycopy(self.radOffset[i], 0, rad_offsets, cnt, self.numChannelsPerDataSet[i])
        cnt += self.numChannelsPerDataSet[i]             
      vals1_20  = jarray.zeros(len(values), 'f')
      vals21_36 = jarray.zeros(len(values), 'f')
      System.arraycopy(values, 0, vals1_20, 0, len(values))
      System.arraycopy(values, 0, vals21_36, 0, len(values))
      
      vals1_20  = Helper.unpack(vals1_20, rad_scales, rad_offsets)
      vals21_36 = Helper.unpack(vals21_36, rad_scales, rad_offsets)
      
      for i in xrange(len(values)):
        if i <= 21:
          vals1_20[i]  = vals1_20[i]
          vals21_36[i] = Float.NaN
        else:
          vals1_20[i] = Float.NaN
          vals21_36[i] = vals21_36[i]
          
      new_vals1_20  = jarray.zeros(len(values), 'f')
      new_vals21_36 = jarray.zeros(len(values), 'f')
      for i in xrange(len(values)):
        new_vals1_20[i]  = vals1_20[self.fileIndexes[i]]
        new_vals21_36[i] = vals21_36[self.fileIndexes[i]]
      return [new_vals1_20,new_vals21_36] 
      
    #- Brightness Temperature for emissive bands, Reflectance for reflective
    elif paramName == "brightnessTemp": 
      rad_scales  = jarray.zeros(self.numObsChannels, 'f')
      rad_offsets = jarray.zeros(self.numObsChannels, 'f')
      ref_scales  = jarray.zeros(self.numObsChannels, 'f')
      ref_offsets = jarray.zeros(self.numObsChannels, 'f')
      cnt1 = 0
      cnt2 = 0
      for i in xrange(self.numObsDataSets):
        if self.reflScale[i] != None:
          System.arraycopy(self.reflScale[i], 0, ref_scales, cnt1, self.numChannelsPerDataSet[i])
          System.arraycopy(self.reflOffset[i], 0, ref_offsets, cnt1, self.numChannelsPerDataSet[i])
          cnt1 += self.numChannelsPerDataSet[i]
        System.arraycopy(self.radScale[i], 0, rad_scales, cnt2, self.numChannelsPerDataSet[i])
        System.arraycopy(self.radOffset[i], 0, rad_offsets, cnt2, self.numChannelsPerDataSet[i])
        cnt2 += self.numChannelsPerDataSet[i] 
            
      vals1_20   = jarray.zeros(len(values), 'f')
      vals21_36  = jarray.zeros(len(values), 'f')
      System.arraycopy(values, 0, vals1_20, 0, len(values))
      System.arraycopy(values, 0, vals21_36, 0, len(values))
      
      vals1_20  = Helper.unpack(vals1_20, ref_scales, ref_offsets)
      vals1_20  = Helper.applyZenithCorrection(vals1_20, self.imageElement, self.imageLine, self.solarZenith, self.gxlength, self.gylength, self.imageDomain)
      vals21_36 = Helper.unpack(vals21_36, rad_scales, rad_offsets)
      
      for i in xrange(len(values)):
        if i <= 21:
          vals1_20[i]  = vals1_20[i]
          vals21_36[i] = Float.NaN
        else:
          vals1_20[i]  = Float.NaN
          vals21_36[i] = vals21_36[i]
      
      for i in xrange(16):
        vals21_36[22+i] = (self.radianceToBrightnessTemp([vals21_36[22+i]], 22+i))[0]
        
      new_vals1_20  = jarray.zeros(len(values), 'f')
      new_vals21_36 = jarray.zeros(len(values), 'f')
      for i in xrange(len(values)):
        new_vals1_20[i]  = vals1_20[self.fileIndexes[i]]
        new_vals21_36[i] = vals21_36[self.fileIndexes[i]]
        
      return [new_vals1_20, new_vals21_36]

  def getSolarZenith(self, name, start, count, stride):
    values = self.reader.getValues(name, start, count, stride)
    return Helper.unpackSolarZenith(values, 0.01, 0, true)
      
  def getImageRange(self, values, paramName, dataset_index, index):
    if paramName == "radiance":
      unit = Parser.parse("W/m^2/micrometer/steradian")
      self.imageRangeUnit = ScaledUnit(1, unit, "W/m^2/\xb5m/ster")
      return Helper.unpack(values, self.radScale[dataset_index][index], self.radOffset[dataset_index][index])
    elif paramName == "brightnessTemp":
      self.imageRangeUnit = Parser.parse("K")
      if dataset_index < 3:
        values = Helper.unpack(values, self.reflScale[dataset_index][index], self.reflOffset[dataset_index][index])
        return Helper.applyZenithCorrection(values, self.solarZenith, self.gxlength, self.gylength, self.imageDomain)
      else:
        vals = Helper.unpack(values, self.radScale[dataset_index][index], self.radOffset[dataset_index][index])
        return self.radianceToBrightnessTemp(vals, (index + 22))
      
  def radianceToBrightnessTempSpec(self, values):
    new_values = jarray.zeros(len(values), 'd')
    for i in xrange(len(values)):
      new_values[i] = (self.radianceToBrightnessTemp([values[i]], self.observationChannels[i]))[0]
    return new_values
    
  def radianceToBrightnessTemp(self, values, channelIndex):
    return Helper.modis_radiance_to_brightnessTemp(self.modis_platform, int(self.band_numbers[channelIndex]), values)
    
  def brightnessTempToRadiance(self, values, wavenumber):
    #Constants are from "The Fundamental Physical Constants",
    #- Cohen, E. R. and B. N. Taylor, Physics Today, August 1993.

    w = wavenumber

    #- Planck constant (Joule second)
    h = 6.6260755e-34

    #- Speed of light in vacuum (meters per second)
    c = 2.9979246e8

    #- Boltzmann constant (Joules per Kelvin)
    k = 1.380658e-23

    #- Derived constants
    c1 = 2.0 * h * c * c
    c2 = (h * c) / k

    #- Convert wavelength to meters
    ws = 1.0e-6 * w

    #- Compute brightness temperature
    nvals = len(values)
    new_values = jarray.zeros(nvals, 'f')
    for i in xrange(nvals):
      t = values[i]
      new_values[i] = (1.0e-6*(c1/(ws*ws*ws*ws*ws)))/(Math.exp(c2/(ws * t)) - 1.0)
    return new_values

  #-- should use units and look in configuration file?  
  def getObsValidRange(self):
    if self.paramName   == "radiance":
      return [[0,500],[0, 10]]
    elif self.paramName == "brightnessTemp":
      return [[0,1],[100, 320]]
  
  def channelIndexToDataSetIndex(self, channelIndex):
    idx = self.fileIndexes[channelIndex] #- get the unsorted index
    if idx <= 1:
      return 0
    elif idx <= 6:
      return 1
    elif idx <= 21:
      return 2
    else:
      return 3
    
  def clone(self):
    return ModisData(self.filename,self.reader,self.readerTable,master=self)     
    
  def channelIndexToFileIndex(self, datasetIndex, channelIndex):
    idx = self.fileIndexes[channelIndex]
    if datasetIndex == 0:
      return idx
    else:
      a = 0
      for i in xrange(datasetIndex):
        a += self.numChannelsPerDataSet[i]   
      return idx - a
        
  def makeRangeTypes(self):
    self.spectrumRangeTypeDict = {
      "radiance": 
        RealTupleType(RealType.getRealType("Radiance_1-19_26"), RealType.getRealType("Radiance_20-36")),
      "brightnessTemp": 
        RealTupleType(RealType.getRealType("Reflectance"), RealType.getRealType("BrightnessTemp"))}
        
    self.imageRangeTypeDict = {
      "radiance": RealType.getRealType("Radiance"),
      "brightnessTemp": RealType.getRealType("Refl/BT")}
        
    self.rangeTypes = []
    self.rangeTypes.append(RealType.getRealType("Radiance_1-20"))
    self.rangeTypes.append(RealType.getRealType("Radiance_21-36"))
    self.rangeTypes.append(RealType.getRealType("Reflectance"))
    self.rangeTypes.append(RealType.getRealType("BrightnessTemp"))
    
  def makeSpectrumDomainType(self):
    micrometer = ScaledUnit(0.000001, CommonUnit.meter, "\xb5m")
    self.obsChannelsType = RealType.getRealType("wavelength", micrometer)
    
  def makeSpectrumRangeType(self, paramName):
    if paramName == "radiance":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[0], self.rangeTypes[1])
    elif paramName == "brightnessTemp":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[2], self.rangeTypes[3])
      
  def makeImageRangeType(self, paramName):
    if paramName == "radiance":
      self.imageRangeType = RealType.getRealType("radiance")
    elif paramName == "brightnessTemp":
      self.imageRangeType = RealType.getRealType("brightnessTemp")
      
  def bandNameToValue(self, text):
    return self.observationChannels[list(self.fileIndexes).index((self.bandDict[text])[0])]
    
  def bandValueToName(self, value):
    idx = self.observationDomain.valueToIndex([[value]])[0]
    return self.band_numbers[self.fileIndexes[idx]]
#------------------------------------------------------------------------------------------

class MsgData(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)

  def getObservationChannels(self):
    self.fileIndexes = None
    self.observationChannels = [0.6, 0.8, 1.6, 3.9, 6.2, 7.3, 8.7, 9.7, 10.8, 12.0, 13.4]
    self.bandDict = {}
    self.band_numbers = ['1','2','3','4','5','6','7','8','9','10','11']
    self.fileIndexes = QuickSort.sort(self.observationChannels)
    cnt = 0
    for i in xrange(len(self.observationChannels)):
      band_number = self.band_numbers[i]
      center_wn = self.observationChannels[i]
      self.bandDict.update({band_number : [cnt,center_wn]})
      self.bandDict.update({center_wn : band_number})
      cnt += 1

    return self.observationChannels

  """
  def channelIndexToDataSetIndex(self, channelIndex):
    if channelIndex <= 2:
      return 0
    else:
      return 1

  def channelIndexToFileIndex(self, datasetIndex, channelIndex):
    if datasetIndex == 0:
      return channelIndex
    else:
      a = 0
      for i in xrange(datasetIndex):
        a += self.numChannelsPerDataSet[i]
      return channelIndex - a
  """

  def makeRangeTypes(self):
    self.spectrumRangeTypeDict = {
      "radiance":
        RealTupleType(RealType.getRealType("Radiance_1-3"), RealType.getRealType("Radiance_4-11")),
      "brightnessTemp":
        RealTupleType(RealType.getRealType("Reflectance"), RealType.getRealType("BrightnessTemp"))}

    self.imageRangeTypeDict = {
      "radiance": RealType.getRealType("Radiance"),
      "brightnessTemp": RealType.getRealType("Refl/BT")}

    self.rangeTypes = []
    self.rangeTypes.append(RealType.getRealType("Radiance_1-3"))
    self.rangeTypes.append(RealType.getRealType("Radiance_4-11"))
    self.rangeTypes.append(RealType.getRealType("Reflectance"))
    self.rangeTypes.append(RealType.getRealType("BrightnessTemp"))

  def makeSpectrumDomainType(self):
    micrometer = ScaledUnit(0.000001, CommonUnit.meter, "\xb5m")
    self.obsChannelsType = RealType.getRealType("wavelength", micrometer)

  def makeSpectrumRangeType(self, paramName):
    if paramName == "radiance":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[0], self.rangeTypes[1])
    elif paramName == "brightnessTemp":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[2], self.rangeTypes[3])

  def makeImageRangeType(self, paramName):
    if paramName == "radiance":
      self.imageRangeType = RealType.getRealType("radiance")
    elif paramName == "brightnessTemp":
      self.imageRangeType = RealType.getRealType("brightnessTemp")

  def getSpectrumRange(self,values, paramName):
    vals1_3  = jarray.zeros(len(values), 'f')
    vals4_11 = jarray.zeros(len(values), 'f')

    for i in xrange(len(values)):
      vals1_3[i]  = Float.NaN
      vals4_11[i] = Float.NaN

    if paramName == "radiance":
      for i in xrange(len(values)):
        if i <= 2:
          vals1_3[i]  = values[i]
        else:
          vals4_11[i] = values[i]
    elif paramName == "brightnessTemp":
      tmp = Helper.msg_radiance_to_reflectance([0,1,2], values[0:3])
      for i in xrange(len(tmp)):
        vals1_3[i] = tmp[i]

      tmp = Helper.msg_radiance_to_brightnessTemperature([0,1,2,3,4,5,6,7], values[3:11])
      for i in xrange(len(tmp)):
        vals4_11[i+3] = tmp[i]

    return [vals1_3,vals4_11]

  def getImageRange(self, values, paramName, dataset_index, index):
    if paramName == "radiance":
      unit = Parser.parse("mW/m^2/cm^-1/steradian")
      self.imageRangeUnit = ScaledUnit(1, unit, "mW/m^2/\xb5m/ster")
      return values
    elif paramName == "brightnessTemp":
      self.imageRangeUnit = Parser.parse("K")
      if index < 3:
        return self.radianceToReflectance(values, index)
      else:
        return self.radianceToBrightnessTemp(values, index-3)

  def radianceToReflectance(self, values, index):
    return Helper.msg_radiance_to_reflectance(index, values)

  def radianceToBrightnessTemp(self, values, channelIndex):
    return Helper.msg_radiance_to_brightnessTemperature(channelIndex, values)

  def bandNameToValue(self, text):
    return (self.bandDict[text])[1]

  def bandValueToName(self, value):
    idx = self.observationDomain.valueToIndex([[value]])[0]
    return self.band_numbers[idx]

  def getObsValidRange(self):
    if self.paramName   == "radiance":
      return [[0,10],[0, 10]]
    elif self.paramName == "brightnessTemp":
      return [[0,1],[100, 320]]

  def getInitialSubset(self):
    return Subset(None,None,None,None,step=10)

  def clone(self):
    return MsgData(self.filename,self.reader,self.readerTable,master=self)
    
#------------------------------------------------------------------------------------------------ 
#-- break this out into a separate module     
class AirsData(MultiSpectralData):
  def __init__(self,filename,reader,readerTable,channelPropertyFile,master=None):
    MultiSpectralData.__init__(self,filename,reader,readerTable,channelPropertyFile,master)
        
    if filename.find("http:") == -1:
      attr = reader.getGlobalAttr(3)
      ss = str(attr)

      i = ss.find('RANGEBEGINNINGDATE')
      ii = ss.find('END_OBJECT', i)
      new_ss = attr.substring(i,ii)
      ssa=new_ss.splitlines()
      self.date_str = (((ssa[2].split('=')))[1].split('"'))[1]

      i = ss.find('RANGEBEGINNINGTIME')
      ii = ss.find('END_OBJECT', i)
      new_ss = attr.substring(i,ii)
      ssa=new_ss.splitlines()
      self.time_str = (((ssa[2].split('=')))[1].split('"'))[1]
    else:
      attrs_list = reader.getAllGlobalAttrs()
      ii = -1
      for i in xrange(len(attrs_list[0])):
        if attrs_list[0][i] == "coremetadata.coremetadata.INVENTORYMETADATA.RANGEDATETIME.VALUE":
          ii = i
          break
      self.date_str = attrs_list[1][ii]
      self.time_str = attrs_list[1][ii+2]

    if master != None:
      self.rad_quality = master.rad_quality
      self.date_str    = master.date_str
      self.time_str    = master.time_str

  #- override
  def getObservationChannels(self):
    #if self.channelPropertyFile == None:
    #  return MultiSpectralData.getObservationChannels(self)
      
    ff  = open(hydra_path+'/ancillary/airs/L2.chan_prop.2003.11.19.v6.6.9.anc', 'r')
    lines    = ff.readlines()
    nlines   = len(lines)
    channels         = []
    good_indexes     = []
    self.rad_quality = []
    cnt = 0
    for i in xrange(nlines):
      if lines[i].startswith('!'):
        pass
      else:
        fields = lines[i].split()
        srf_centroid_freq = fields[1]
        rq = fields[11]
        l2_ignore = fields[12]
        channels.append(float(srf_centroid_freq))
        good_indexes.append(cnt)
        self.rad_quality.append(int(l2_ignore))
        cnt += 1
    self.fileIndexes = QuickSort.sort(channels)
    new_channels     = jarray.zeros(len(channels), 'f')
    
    for i in xrange(len(new_channels)):
      new_channels[i] = channels[self.fileIndexes[i]]
    return new_channels
 
  def makeSpectrumDomainType(self):
    centimeter = ScaledUnit(0.01, CommonUnit.meter, "cm")
    inv_cm = centimeter.pow(-1)
    inv_cm = ScaledUnit(1, inv_cm, "cm^-1")
    self.obsChannelsType = getRealType("wavenumber", inv_cm)
    
  def getSpectrumRange(self, values, paramName):
    #first, apply QC flags from channel property file
    nn = len(values)
    new_values = jarray.zeros(len(values), 'f')
    for i in xrange(nn):
      val = values[self.fileIndexes[i]]
      rq  = self.rad_quality[self.fileIndexes[i]]
      if rq == 0:
        new_values[i] = val
      else:
        new_values[i] = Float.NaN
    
    if self.native_param == paramName:
      return new_values
    elif paramName == "brightnessTemp":
      return self.radianceToBrightnessTempSpectrum(new_values)
    elif paramName == "radiance":
      return self.brightnessTempToRadiance(new_values,channelIndex)
    
        
  #-override  
  def channelIndexToFileIndex(self, dataset_index, idx):
    return self.fileIndexes[idx]
       
  #- override
  def getObsValidRange(self):
    if self.paramName   == "radiance":
      return [0, 150]
    elif self.paramName == "brightnessTemp":
      return [180, 320]
  
  #- override  
  def initMissing(self, longitudes, latitudes, values):
    class Missing:
      def __init__(self, longitudes, latitudes, values, xlength, ylength, invert):
        self.longitudes = longitudes
        self.latitudes  = latitudes
        self.good_lines = []
        
        nlines = xlength
        nelems = ylength
        self.invert = invert
        if invert == true:
          nlines = ylength
          nelems = xlength
          
        for i in xrange(nlines):
          cnt = 0
          for j in xrange(nelems):
            if invert == true:
              ii = j + i*nelems
            else:
              ii = i + j*nlines
            lat = latitudes[ii]
            lon = longitudes[ii]
            val = values[ii]
            if ((lon <= 180 and lon >= -180) and (lat <= 90 and lat >= -90) and val > -9000):
              cnt += 1
          if cnt == nelems:    
            self.good_lines.append(i)
        self.n_good_lines = len(self.good_lines)
        self.nelems = nelems
        
        self.xlength = self.n_good_lines
        self.ylength = self.nelems
        if invert == true:
          self.ylength = self.n_good_lines
          self.xlength = self.nelems
                      
      def dataToSegment(self, line, element):
        return self.good_lines[line], element
        
      def removeMissing(self, values):
        new_values = jarray.zeros(self.n_good_lines*self.nelems, 'f')
        for i in xrange(self.n_good_lines):
          for j in xrange(self.nelems):
            if self.invert == true:
              ii  = j + i*self.nelems
              iii = j + self.good_lines[i]*self.nelems
            else:
              ii  = i + j*self.n_good_lines
              iii = self.good_lines[i] + j*self.nlines
            new_values[ii] = values[iii]  
        return new_values
        
      def getLengths(self):
        return self.xlength, self.ylength
    
    m = Missing(longitudes, latitudes, values, self.xlength, self.ylength, self.invert)
    return m

  def clone(self):
    return AirsData(self.filename,self.reader,self.readerTable,self.channelPropertyFile,master=self)
    
class Level2Data(MultiSpectralData):
  def __init__(self,filename,reader,readerTable,master=None):
    MultiSpectralData.__init__(self,filename,reader,readerTable,master)
    if master == None:
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      self.getScaleOffset()
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
  
    self.pressureLevels=self.readPressureLevels()
    self.isProfile = self.determineIfProfile()
    self.pressureType=RealType.getRealType("pressure_dummy")

    self.logPressType = RealType.getRealType('pressure')
    reference = RealTupleType(self.logPressType)
    logCS = LogCoordinateSystem(reference, 10)
    range = RealTupleType(self.pressureType, logCS, None)
    self.pressureType = range

    self.rangeTypes.append(self.pressureType)
    ###self.levelValue=0 # this is the surface
    self.makeValidRange()

    self.observationDomain = Gridded1DSet(self.pressureType, [self.pressureLevels], len(self.pressureLevels))

    self.initPressure = 500
    self.levelValue   = self.observationDomain.valueToIndex([[self.initPressure],])[0]
    self.initLevelIndex = self.levelValue
  
  def readPressureLevels(self):
    if self.instrumentName == "AIRSUWLevel2":
      ff = open(hydra_path+'/ancillary/airs/uwLevel2PressureLevels.txt', 'r')
    elif self.instrumentName == "AIRSLevel2":
      ff = open(hydra_path+'/ancillary/airs/jplLevel2PressureLevels.txt', 'r')
    elif self.instrumentName == "MODISUWLevel2":
      ff = open(hydra_path+'/ancillary/modis/uwLevel2PressureLevels.txt', 'r')
    elif self.instrumentName == "MODISLevel2":
      ff = open(hydra_path+'/ancillary/modis/uwLevel2PressureLevels.txt', 'r')
    lines=ff.readlines()
    nlines=len(lines)
    pLevels=[]
    for i in xrange(nlines):
      pLevels.append(float(lines[i]))

    return pLevels  
             
  def determineIfProfile(self):
    if self.paramName == "Air Temperature Profiles":
      return true
    elif self.paramName == "Surface Pressure":
      return false
    elif self.paramName == "Surface Air Temperature" or self.paramName == "Surface Temperature" or self.paramName == "Surface/Surface Air Temp Difference" :
      return false
    elif self.paramName == "Precipitable Water":
      return false
    elif self.paramName == "Total Ozone":
      return false
    elif self.paramName == "Ozone Profiles":
      return true
    elif self.paramName == "Water Vapor Profiles":
      return true
  
  def extractValues(self,dataToExtract):
    dataToExtract = Helper.missingToNaN(dataToExtract, -9990.0)
    return dataToExtract
 
  def makeRangeTypes(self):
    # Create a dictionary corresponding to the names in the Level 2 Files
    self.rangeTypes = []
    paramNames = self.rangeLongName
    self.imageRangeTypeDict={}
    for i in xrange(self.numObsDataSets):
        self.rangeTypes.append(RealType.getRealType(self.observationName[i]))
        self.imageRangeTypeDict[paramNames[i]]=RealType.getRealType(self.observationName[i])
        
    self.spectrumRangeTypeDict    = None

  def makeValidRange(self):
    self.validRangeDict={}
    if self.rangeValidMin == None:
      self.validRangeDict["Precipitable Water"] = [0.0,100.0]
      self.validRangeDict["Total Ozone"] = [0.0,500.0]
      self.validRangeDict["Surface Temperature"] = [180.0,400.0]
      self.validRangeDict["Surface AIR Temperature"] = [180.0,400.0]
      self.validRangeDict["Surface/Surface Air Temp Difference"] = [-40.0,40.0]
      self.validRangeDict["Air Temperature Profiles"] = [180.0,320.0]
      self.validRangeDict["Surface Pressure"] = [600.0,1050.0]
      self.validRangeDict["Ozone Profiles"] = [0.0,1.0]
      self.validRangeDict["Water Vapor Profiles"] = [0.0,30.0]
    else:
      for i in xrange(len(self.rangeLongName)):
        lo = int(self.rangeValidMin[i])
        hi = int(self.rangeValidMax[i])
        self.validRangeDict[self.rangeLongName[i]] = [lo,hi] 
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None

  def getNativeParam(self):
    native_param="Air Temperature Profiles"
    return native_param

  def setDefaultParam(self):
    return self.native_param
    
  def getPressureRange(self):
     lenField=len(self.pressureLevels)
     return [self.pressureLevels[0]+50,self.pressureLevels[lenField-1]]
     
  def getObsValidRange(self):
    return self.validRangeDict[self.paramName]
      
  def getImageRange(self, values, paramIndex):
    # Level 2 AIRS Products do not have a scale and offset defined
    if self.paramName == "Air Temperature Profiles":
      self.imageRangeUnit = Parser.parse("K")
    elif self.paramName == "Surface Pressure":
      self.imageRangeUnit = Parser.parse("hPa")
    elif self.paramName == "Surface Air Temperature" or self.paramName == "Surface Temperature" or self.paramName == "Surface/Surface Air Temp Difference":
      self.imageRangeUnit = Parser.parse("K")
    elif self.paramName == "Precipitable Water":
      self.imageRangeUnit = Parser.parse("cm")
    elif self.paramName == "Total Ozone":
      db=BaseUnit.addBaseUnit("Dobsons","db","db")
      self.imageRangeUnit = Parser.parse(db.toString())
    elif self.paramName == "Ozone Profiles":
      db=BaseUnit.addBaseUnit("ppmv","ppmv","ppmv")
    elif self.paramName == "Water Vapor Profiles":
      self.imageRangeUnit = Parser.parse("g/kg")

    valid = self.validRangeDict[self.rangeLongName[paramIndex]]
    if self.paramScale[paramIndex] != None and self.paramOffset[paramIndex] != None:
      values = Helper.unpack(values, self.paramScale[paramIndex][0], self.paramOffset[paramIndex][0], valid[0], valid[1])
    else:
      values = Helper.missingToNaN(values, valid[0], valid[1])
    return values

  def getProfileRange(self, values, paramIndex):
    valid = self.validRangeDict[self.rangeLongName[paramIndex]]
    if self.paramScale[paramIndex] != None and self.paramOffset[paramIndex] != None:
      values = Helper.unpack(values, self.paramScale[paramIndex][0], self.paramOffset[paramIndex][0], valid[0], valid[1])
    return values
  
  def getObservations(self, line, element, copy=false):
    # get the values at a point - for AIRS Level 2, this is valuable for getting profiles
    if (self.invert == true):
      tmp     = element
      element = line
      line    = tmp

    line, element = self.missingObs.dataToSegment(line,element)
      
    if (self.subset != None):      
      line    *= self.step
      element *= self.step
      if (self.invert == true):
        line     += self.subset.start_y
        element  += self.subset.start_x
      else:
        line     += self.subset.start_x
        element  += self.subset.start_y
    else:
      line    *= self.step
      element *= self.step
      
    ll = int(line + (self.trackSegments[self.currentSegmentIndex*2]-1)) #convert one to zero based

    # access the data in the file
    if len(self.obsDimList) == 3:
      vals = jarray.zeros(len(self.pressureLevels), 'f')
      start  = [0,element+self.geoOffset,ll+self.geoOffset]
      count  = [len(self.pressureLevels),1,1]
      nstart = [None, None, None]
      ncount = [None, None, None]
    elif len(self.obsDimList) == 2:
      # only 1 element, but make an array out of it anyway
      vals = jarray.zeros(1, 'f')
      start  = [element+self.geoOffset,ll+self.geoOffset]
      count  = [1,1]
      nstart = [None, None]
      ncount = [None, None]
      
    for j in xrange(len(start)):
      nstart[self.obsDimOrder[j]]  = start[j]
      ncount[self.obsDimOrder[j]]  = count[j]
        
    cnt = 0    
    vv = self.reader.getValues(self.fieldToRetrieve, nstart, ncount)      
    for k in xrange(len(vv)):
      if vv[k] != -9999:      
        vals[cnt] = float(vv[k])
      else:
        vals[cnt] = Float.NaN 
      cnt += 1

    vals = self.getProfileRange(vals, self.paramIndex)
              
    #-- make the VisAD data object
    if self.observationField == None or copy == true: 
      obs = FlatField(FunctionType(self.pressureType, self.imageRangeTypeDict[self.paramName]), self.observationDomain)
      obs.setSamples([vals])
      if copy == false:
        self.observationField = obs
      return obs
    else:
      self.observationField.setSamples([vals])
        
    return self.observationField
    
  def getImageSegment(self,segmentIndex=0,subset=None):  
    if self.paramName == None:
      return None
    
    self.fieldToRetrieve=self.imageRangeTypeDict.get(self.paramName).toString()
    # see if difference of fields exists in start-up files
    curField=self.fieldToRetrieve.split("-")
                
    if self.isProfile:
      # some of the AIRS Level 2 fields are 3-dimensional, so the dimlist and dimorder change
      k = self.observationName.index(self.fieldToRetrieve)
      self.obsDimList = [self.obsLevelName[k],self.obsElementName[k],self.obsLineName[k]]
      self.obsDimOrder  = self.reader.getDimOrder(self.fieldToRetrieve,self.obsDimList)
    else:
      k = self.observationName.index(self.fieldToRetrieve)
      self.obsDimList = [self.obsElementName[k],self.obsLineName[k]]
      self.obsDimOrder  = self.reader.getDimOrder(self.fieldToRetrieve,self.obsDimList)

    self.paramIndex = k
            
    self.makeImageRangeType(self.paramName)
    self.currentSegmentIndex = segmentIndex   
    
    segStartLine = int(self.trackSegments[self.currentSegmentIndex*2]-1)  #convert one to zero based
    segEndLine   = int(self.trackSegments[self.currentSegmentIndex*2+1]-1)
    
    if subset != None: #-- subset logic
      self.subset = subset      
      if subset.end_x == None or subset.end_y == None:
        self.startElem = 0
        self.nElements = int((self.numObsElements)/subset.step)
        self.startLine = segStartLine
        self.nLines    = int((int(segEndLine-self.startLine+1))/subset.step)
      else:                 
        t_line  = subset.start_x
        t_elem  = subset.start_y
        n_lines = int((subset.end_x - subset.start_x)/subset.step) + 1
        n_elems = int((subset.end_y - subset.start_y)/subset.step) + 1
        if (self.invert == true):
          t_elem  = subset.start_x
          t_line  = subset.start_y
          n_elems = int((subset.end_x - subset.start_x)/subset.step) + 1
          n_lines = int((subset.end_y - subset.start_y)/subset.step) + 1
        
        self.startElem = t_elem
        self.startLine = segStartLine + t_line      
        self.nElements = n_elems
        self.nLines    = n_lines
      
      self.step      = self.subset.step
           
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    else:
      self.subset    = None
      self.startElem = 0
      self.nElements = self.numObsElements
      self.startLine = segStartLine
      self.nLines    = int(segEndLine-self.startLine+1)
      
      self.step      = 1  
              
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    if (len(self.obsDimList) == 3):
      # read in all the pressure fields 
      if self.levelValue != None:
        start   = [self.levelValue,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
      else:
        start = [0,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        
      count   = [1,self.nElements,self.nLines]
      stride  = [self.step, self.step, self.step]
      nstart  = [None, None, None]
      ncount  = [None, None, None]
      nstride = [None, None, None]
    elif len(self.obsDimList) == 2:
      start   = [self.startElem+self.geoOffset,self.startLine+self.geoOffset]
      count   = [self.nElements,self.nLines]
      stride  = [self.step, self.step]
      nstart  = [None, None]
      ncount  = [None, None]
      nstride = [None, None]
      
    for i in xrange(len(start)):
      nstart[self.obsDimOrder[i]]  = start[i]
      ncount[self.obsDimOrder[i]]  = count[i]
      nstride[self.obsDimOrder[i]] = stride[i]
    
    # make a multidimensional list in case there is more than one field (e.g., TSurfStd-TAirStd difference)
    valueArray=[None] * len(curField)
    for i in xrange(len(curField)):    
      if self.step > 1: #- this hack will go away when Java-Netcdf implements stride reading
        valueArray[i]=self.reader.getValues(curField[i],nstart,ncount,nstride)
      else:
        valueArray[i]=self.reader.getValues(curField[i],nstart,ncount,nstride)
        
    
    # If there is more than one field, we will assume this is a difference (e.g., TSurfStd-TAirStd)
    # could be done more compactly if numarray library installed
    self.z = jarray.zeros(self.nLines*self.nElements,'f')
    if len(curField) > 1:
      for i in xrange(self.nLines * self.nElements):
        self.z[i] = abs(valueArray[0][i] - valueArray[1][i])
    else:
      self.z = valueArray[0]
        
    # Read in geographic data  
    start  =  [self.startElem/self.ratio, self.startLine/self.ratio] 
    count  =  [nGeoElements, nGeoLines]
    stride =  [self.geo_step, self.geo_step]
    
    nstart =  []
    ncount =  []
    nstride = []
    
    for i in xrange(len(start)):
      nstart.append(start[self.geoDimOrder[i]])
      ncount.append(count[self.geoDimOrder[i]])
      nstride.append(stride[self.geoDimOrder[i]])
      
    if self.step > 1:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, nstride)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, nstride)
    else:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, None)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, None)
    
    self.lats = Helper.copyArray(self.lats)
    self.lons = Helper.copyArray(self.lons)

    if self.invert == true:
      self.ylength  = self.nLines
      self.xlength  = self.nElements
      self.gylength = nGeoLines
      self.gxlength = nGeoElements
    else:
      self.xlength  = self.nLines
      self.ylength  = self.nElements
      self.gxlength = nGeoLines
      self.gylength = nGeoElements

      
    self.missingObs = self.initMissing(self.lons, self.lats, self.z)
    self.xlength, self.ylength = self.missingObs.getLengths()    
    self.z    = self.missingObs.removeMissing(self.z)
    self.lons = self.missingObs.removeMissing(self.lons)
    self.lats = self.missingObs.removeMissing(self.lats)

    
    if self.ratio == 1:
      self.gxlength = self.xlength
      self.gylength = self.ylength
     
    try:
      self.gs = Gridded2DSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
              [self.lats, self.lons],
              self.gxlength,self.gylength,None,None,None,0,0)
      self.cs = GridCoordinateSystem(self.gs)
    except:
      self.cs = None
      print "*** Error Making Image CoordinateSystem ***"
        
    self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs,None)
    self.imageDomain  = makeDomain(self.imageDomType,
          0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
          0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      
    self.subsetDomain = makeDomain(self.imageDomType,
          0, (self.xlength-1)*self.step, self.xlength,
          0, (self.ylength-1)*self.step, self.ylength)
          
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    self.imageField        = FlatField(self.imageFunctionType, self.imageDomain)  
    self.z                 = self.getImageRange(self.z, self.paramIndex)
    
    indexes                = jarray.zeros(2, 'i')
    minmax                 = Helper.minmax(self.z, indexes)
    minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
    maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
    self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
    self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
    self.imageField.setSamples([self.z], false)
       
    return self.imageField    
  
  def getSameImageSegment(self, levelIndex):
    # Gets the same segment as previous, but at a different pressure level
    if levelIndex >= 0:
      if len(self.obsDimList) == 3:
        start   = [levelIndex,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count  =  [1,self.nElements,self.nLines]
        stride =  [1, self.step, self.step]
        nstart  = [None, None, None]
        ncount  = [None, None, None]
        nstride = [None, None, None]
      elif len(self.obsDimList) == 2:
        start  =  [self.startElem+self.geoOffset,self.startLine+self.geoOffset]
        count  =  [self.nElements,self.nLines]
        stride =  [self.step, self.step]
        nstart  = [None, None]
        ncount  = [None, None]
        nstride = [None, None]

      for j in xrange(len(start)):
        nstart[self.obsDimOrder[j]]  = start[j]
        ncount[self.obsDimOrder[j]]  = count[j]
        nstride[self.obsDimOrder[j]] = stride[j]
       
      if self.step > 1: 
        self.z = self.reader.getValues(self.fieldToRetrieve,nstart,ncount,nstride)
      else:
        self.z = self.reader.getValues(self.fieldToRetrieve,nstart,ncount)
      
      self.z = self.missingObs.removeMissing(self.z)
      self.z = self.getImageRange(self.z, self.paramIndex)
      
      indexes                = jarray.zeros(2, 'i')
      minmax                 = Helper.minmax(self.z, indexes)
      minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
      maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
      self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
      self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
      
      self.imageField.setSamples([self.z], false)


  #- override
  def initMissing(self, longitudes, latitudes, values):
    class Missing:
      def __init__(self, longitudes, latitudes, values, xlength, ylength, invert):
        self.longitudes = longitudes
        self.latitudes  = latitudes
        self.good_lines = []

        nlines = xlength
        nelems = ylength
        self.invert = invert
        if invert == true:
          nlines = ylength
          nelems = xlength

        for i in xrange(nlines):
          cnt = 0
          for j in xrange(nelems):
            if invert == true:
              ii = j + i*nelems
            else:
              ii = i + j*nlines
            lat = latitudes[ii]
            lon = longitudes[ii]
            val = values[ii]
            if ((lon <= 180 and lon >= -180) and (lat <= 90 and lat >= -90)):
              cnt += 1
          if cnt == nelems:
            self.good_lines.append(i)
        self.n_good_lines = len(self.good_lines)
        self.nelems = nelems

        self.xlength = self.n_good_lines
        self.ylength = self.nelems
        if invert == true:
          self.ylength = self.n_good_lines
          self.xlength = self.nelems

      def dataToSegment(self, line, element):
        return self.good_lines[line], element

      def removeMissing(self, values):
        new_values = jarray.zeros(self.n_good_lines*self.nelems, 'f')
        for i in xrange(self.n_good_lines):
          for j in xrange(self.nelems):
            if self.invert == true:
              ii  = j + i*self.nelems
              iii = j + self.good_lines[i]*self.nelems
            else:
              ii  = i + j*self.n_good_lines
              iii = self.good_lines[i] + j*self.nlines
            new_values[ii] = values[iii]
        return new_values

      def getLengths(self):
        return self.xlength, self.ylength

    m = Missing(longitudes, latitudes, values, self.xlength, self.ylength, self.invert)
    return m
        
  def clone(self):
    return Level2Data(self.filename,self.reader,self.readerTable,master=self)

#- MODIS Airborne Simulator    
class MasData(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in same directory as L1B
    if master == None:        
      self.getScaleOffset()
    else:
      self.radScale       = master.radScale
      
  def setDefaultParam(self):
    return "radiance"

  def getObservationUnit(self):
    self.observationUnit.append(Parser.parse("W/m^2/steradian/micrometer"))

  def getScaleOffset(self):
    self.radScale, dum = self.reader.getObsScaleOffset(self.observationName[0], "scale_factor", "offset")
        
  def getImageRange(self, values, paramName, dataset_index, index):
    if paramName == "radiance":
      unit = Parser.parse("W/m^2/steradian/micrometer")
      self.imageRangeUnit = ScaledUnit(1, unit, "W/m^2/ster/\xb5m")
      return Helper.unpack(values, self.radScale[index], 0)
    elif paramName == "brightnessTemp":
      self.imageRangeUnit = Parser.parse("K")
      vals = Helper.unpack(values, self.radScale[index], 0)
      return self.radianceToBrightnessTemp(vals, index)
      
  def clone(self):
    return MasData(self.filename,self.reader,self.readerTable,master=self)  
   
# MODIS Cloud Mask
class ModisCloudMaskData(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in the same directory as the MOD35 Product
    if master == None:
      if filename.find("mod35") != -1:
        geo_filename = filename.replace("mod35", "geo")
      elif filename.find("MOD35") != -1:
        geo_filename = filename.replace("MOD35", "MOD03")
      elif filename.find("MYD35") != -1:
        geo_filename = filename.replace("MYD35", "MYD03")
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"
      self.geoOffset = 2
    
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
      self.geoOffset      = master.geoOffset

  def makeTrackSegments(self):
    if self.numObsLines < 2600:
      self.numberSegments = 1
      self.trackSegments = [1, self.numObsLines]
      return
    self.numberSegments = 1
    self.trackSegments = [401, self.numObsLines-100]

  
  # overwrite
  def extractValues(self,dataToExtract,cloudMaskBits,qaDataToExtract,qaBits):   
    # get bit at specified position
    def getBitAt(byteVal,index):    
      # negative integers in Java represented in 2's complement
      # hdf integers written as 2's complement and big-endian
      #b=byteVal & 0x0f
      #mask = 1 << index
      #print b, mask, b & mask
      #if b & mask == mask:
      #   return 1
      #else:
      #   return 0
      return (byteVal >> index) & 1
    
    outMask=jarray.zeros(len(dataToExtract),'f')
       
    for i in xrange(len(dataToExtract)):
      bitVal=[]      
      for j in xrange(len(cloudMaskBits)):
        bitVal.append(getBitAt(dataToExtract[i],int(cloudMaskBits[j])))
             
      # This is true only for Unobstructed FOV Quality Flag and Land or Water Path
      # no quality assurance for these as well
      if len(bitVal) > 1:
        test1=bitVal[1] and bitVal[0]
        test2=bitVal[1] and (not bitVal[0])
        test3=not bitVal[1] and bitVal[0]
        test4=not bitVal[1] and (not bitVal[0])
        if test1:
          outMask[i]=0
        elif test2:
          outMask[i]=1
        elif test3:
          outMask[i]=2
        elif test4:
          outMask[i]=3
      else:
        if qaBits != None:
          qaVal = getBitAt(qaDataToExtract[i],int(qaBits))
          testNotApplied = not(qaVal)
          testAppliedCloudFound = qaVal and not(bitVal[0])
          testAppliedCloudNotFound = qaVal and bitVal[0]
          if testNotApplied:
            outMask[i] = 4
          elif testAppliedCloudFound:
            outMask[i] = 5
          elif testAppliedCloudNotFound:
            outMask[i] = 6
          # adjust depending on what cloud mask byte 0 field selected
          # maximum number of cloud mask byte 0 fields is 4
          outMask[i]=outMask[i] - (4 - self.cloudMaskByte0SampleLen/2)  
        else:
          outMask[i]=bitVal[0]
            
    return outMask
    
  def makeRangeTypes(self):  
    # Cloud mask is a bit different as it contains various quality assurance flags
    self.paramNames=["Cloud Mask","Quality Assurance"]
    self.imageRangeTypeDict={}
    self.rangeTypes=[]
    for i in xrange(len(self.paramNames)):
        self.rangeTypes.append(RealType.getRealType(self.paramNames[i]))
        self.imageRangeTypeDict[self.paramNames[i]]=RealType.getRealType(self.observationName[i])
    self.spectrumRangeTypeDict    = None
    
    self.testDict={}
    self.testNames=[]
    testFile = open(hydra_path+'/ancillary/modis/cloudMaskTests.txt','r')
    lines = testFile.readlines()
    nlines = len(lines)
    count = 0
    for i in xrange(nlines):
      if lines[i].startswith('!'):
        pass
      else:
        fields        = lines[i].strip().split(',')
        testName      = fields[0]
        cloudMaskByte = fields[1]
        cloudMaskBit  = fields[2]
        qaByte        = fields[3]
        if qaByte == '':
            qaByte = None
        qaBit         = fields[4]
        if qaBit == '':
           qaBit = None
        self.testDict.update({testName : [count,cloudMaskByte,cloudMaskBit,qaByte,qaBit]})
        self.testNames.append(testName)
        count += 1
    
    self.testRangeTypes=[]
    self.testRangeTypeDict={}
    for i in xrange(len(self.testNames)):
        self.testRangeTypes.append(RealType.getRealType(self.testNames[i]))
        self.testRangeTypeDict[self.testNames[i]]=RealType.getRealTypeByName(self.testNames[i])
#  def makeRangeTypes(self):
#    self.rangeTypes = []
#    self.rangeTypes.append(RealType.getRealType("cloudMask"))    
#    self.imageRangeTypeDict = {"cloudMask" : RealType.getRealType("cloudMask")}
#    self.spectrumRangeTypeDict    = None

  def makeImageRangeType(self, paramName):
    self.imageRangeType = self.imageRangeTypeDict["Cloud Mask"] 
    self.testRangeType  = self.testRangeTypeDict[paramName] 
  
  def getImageRange(self, values):
    # Cloud Mask Bit Values are unitless      
    self.imageRangeUnit = None
    return values
    
  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
     return [0, 3]
    
  def getNativeParam(self):
    native_param = "Unobstructed_FOV_Quality_Flag"  # actually the native test parameter which is the final results of the cloud mask tests
    return native_param

  def setDefaultParam(self):
    return self.native_param
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None
    
  def getImageSegment(self, segmentIndex=0, subset=None):
    byteIndex = self.byteIndex
    cloudMaskBits = self.testMaskInfo[2].split(':')
    qaByteIndex=self.testMaskInfo[3]
    qaBits=self.testMaskInfo[4]
    
    self.currentSegmentIndex = segmentIndex   
    self.makeImageRangeType(self.paramName)
    
    segStartLine = int(self.trackSegments[self.currentSegmentIndex*2]-1)  #convert one to zero based
    segEndLine   = int(self.trackSegments[self.currentSegmentIndex*2+1]-1)
    
    if subset != None: #-- subset logic
      self.subset = subset      
      if subset.end_x == None or subset.end_y == None:
        self.startElem = 0
        self.nElements = int((self.numObsElements)/subset.step)
        self.startLine = segStartLine
        self.nLines    = int((int(segEndLine-self.startLine+1))/subset.step)
      else:                 
        t_line  = subset.start_x
        t_elem  = subset.start_y
        n_lines = int((subset.end_x - subset.start_x)/subset.step) + 1
        n_elems = int((subset.end_y - subset.start_y)/subset.step) + 1
        if (self.invert == true):
          t_elem  = subset.start_x
          t_line  = subset.start_y
          n_elems = int((subset.end_x - subset.start_x)/subset.step) + 1
          n_lines = int((subset.end_y - subset.start_y)/subset.step) + 1
        
        self.startElem = t_elem
        self.startLine = segStartLine + t_line      
        self.nElements = n_elems
        self.nLines    = n_lines
      
      self.step      = self.subset.step
           
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
        
    else:
      self.subset    = None
      self.startElem = 0
      self.nElements = self.numObsElements
      self.startLine = segStartLine
      self.nLines    = int(segEndLine-self.startLine+1)
      
      self.step      = 1  
      
      if float(self.ratio)/float(self.step) <= 1:
        self.geo_step = int(1.0/(float(self.ratio)/float(self.step)))
        nGeoLines     = self.nLines
        nGeoElements  = self.nElements
      else:
        self.geo_step = 1
        nGeoLines     = int(float(self.nLines)/float(self.ratio)/float(self.step))
        nGeoElements  = int(float(self.nElements)/float(self.ratio)/float(self.step))
   
   # cloud mask byte arranged as Byte, Cell_Along, Cell_Across     
   # qa arranged as Cell_Along, Cell_Across, QA_Byte (for some reason)
    start   = [byteIndex,self.startElem+self.geoOffset,self.startLine+self.geoOffset]
    count   = [1,self.nElements,self.nLines]
    stride  = [1, self.step, self.step]
    nstart  = [None, None, None]
    ncount  = [None, None, None]
    nstride = [None, None, None]
    
    if qaByteIndex != None:
      qaStart   = [self.startElem+self.geoOffset,self.startLine+self.geoOffset,int(qaByteIndex)]
      qaCount   = [self.nElements,self.nLines,1]
      qaStride  = [self.step, self.step,1]
      qaNStart  = [None, None, None]
      qaNCount  = [None, None, None]
      qaNStride = [None, None, None]
      # get the QA Dimension order (different from Cloud Mask Order)
      self.qaDimList = [self.obsElementName[0],self.obsLineName[0],self.channelIndexName[1]]
      self.qaDimOrder          = self.reader.getDimOrder(self.observationName[1],self.qaDimList)

    for i in xrange(len(start)):
      nstart[self.obsDimOrder[i]]  = start[i]
      ncount[self.obsDimOrder[i]]  = count[i]
      nstride[self.obsDimOrder[i]] = stride[i]
      if qaByteIndex != None:
        qaNStart[self.qaDimOrder[i]] = qaStart[i]
        qaNCount[self.qaDimOrder[i]] = qaCount[i]
        qaNStride[self.qaDimOrder[i]] = qaStride[i]
      
    if self.step > 1: #- this hack will go away when Java-Netcdf implements stride reading
      self.maskValues = self.reader.getValues(self.observationName[0], nstart, ncount, nstride)
      if qaByteIndex != None:
        self.qaValues = self.reader.getValues(self.observationName[1], qaNStart, qaNCount, qaNStride)
      else:
        self.qaValues = None
    else:
      self.maskValues = self.reader.getValues(self.observationName[0], nstart, ncount, None)
      if qaByteIndex != None:
        self.qaValues = self.reader.getValues(self.observationName[1], qaNStart, qaNCount, None)
      else:
        self.qaValues = None
          
#   call used to extract data from the cloud mask using QA
    self.z=self.extractValues(self.maskValues,cloudMaskBits,self.qaValues,qaBits)
            
    # Read in geographic data  
    start  =  [self.startElem/self.ratio, self.startLine/self.ratio] 
    count  =  [nGeoElements, nGeoLines]
    stride =  [self.geo_step, self.geo_step]
    nstart =  []
    ncount =  []
    nstride = []
    
    for i in xrange(len(start)):
      nstart.append(start[self.geoDimOrder[i]])
      ncount.append(count[self.geoDimOrder[i]])
      nstride.append(stride[self.geoDimOrder[i]])
      
    if self.step > 1:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, nstride)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, nstride)
    else:
      self.lons = self.geo_reader.getValues(self.obsLongitudeName, nstart, ncount, None)
      self.lats = self.geo_reader.getValues(self.obsLatitudeName, nstart, ncount, None)
      
    if self.invert == true:
      self.ylength  = self.nLines
      self.xlength  = self.nElements
      self.gylength = nGeoLines
      self.gxlength = nGeoElements
    else:
      self.xlength  = self.nLines
      self.ylength  = self.nElements
      self.gxlength = nGeoLines
      self.gylength = nGeoElements

    self.missingObs = self.initMissing(self.lons, self.lats, self.z)
    self.xlength, self.ylength = self.missingObs.getLengths()    
    self.z    = self.missingObs.removeMissing(self.z)
    self.lons = self.missingObs.removeMissing(self.lons)
    self.lats = self.missingObs.removeMissing(self.lats)
    
    if self.ratio == 1:
      self.gxlength = self.xlength
      self.gylength = self.ylength
     
    try:
      self.gs = Gridded2DSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
              [self.lats, self.lons],
              self.gxlength,self.gylength,None,None,None,0,0)
      self.cs = GridCoordinateSystem(self.gs)
    except:
      try:
         self.gs = Gridded2DDoubleSet(RealTupleType(self.obsLatitudeType,self.obsLongitudeType),
                                      [self.lats, self.lons],
                                      self.gxlength,self.gylength,None,None,None,0)
      except Exception, e:
        print e
        self.cs = None
        print "*** Error Making Image CoordinateSystem ***"
        
    self.cs = GridCoordinateSystem(self.gs)

    self.imageDomType = RealTupleType(self.spatialTupleArray[0], self.spatialTupleArray[1], self.cs,None)
    self.imageDomain  = makeDomain(self.imageDomType,
          0,((self.xlength)/(self.xlength/self.gxlength))-1,self.xlength,
          0,((self.ylength)/(self.ylength/self.gylength))-1,self.ylength)
      
      
    self.subsetDomain = makeDomain(self.imageDomType,
          0, (self.xlength-1)*self.step, self.xlength,
          0, (self.ylength-1)*self.step, self.ylength)
          
    self.imageFunctionType = FunctionType(self.imageDomType, self.imageRangeType)
    self.imageField        = FlatField(self.imageFunctionType, self.imageDomain)  
    self.z                 = self.getImageRange(self.z)
        
    indexes                = jarray.zeros(2, 'i')
    minmax                 = Helper.minmax(self.z, indexes)
    minloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[0]]))
    maxloc = self.cs.toReference(self.imageDomain.indexToValue([indexes[1]]))
    self.obsMin = [minloc[0][0], minloc[1][0], minmax[0]]
    self.obsMax = [maxloc[0][0], maxloc[1][0], minmax[1]]
    self.imageField.setSamples([self.z], false)
       
    return self.imageField
      
  def clone(self):
    return ModisCloudMaskData(self.filename,self.reader,self.readerTable,master=self)
      
# MODIS Cloud Phase (for now)
class Modis06Data(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in the same directory as the MOD06 Product
    if master == None:
      if filename.find("mod06") != -1:
        geo_filename = filename.replace("mod06", "geo")
      elif filename.find("MOD06") != -1:
        geo_filename = filename.replace("MOD06", "MOD03")
      elif filename.find("MYD06") != -1:
        geo_filename = filename.replace("MYD06", "MYD03")
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"

      self.numObsElements_radii = self.numObsElements*5
      self.numObsLines_radii = self.numObsLines*5
      self.numObsElements_save = self.numObsElements
      self.numObsLines_save = self.numObsLines
    
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
      self.numObsElements_radii = master.numObsElements_radii
      self.numObsLines_radii = master.numObsLines_radii
      self.numObsElements_save = master.numObsElements_save
      self.numObsLines_save = master.numObsLines_save
  
  def getScaleOffset(self):
    self.paramScale   = []
    self.paramOffset  = []
    
    self.fieldToRetrieve=self.imageRangeTypeDict.get(self.paramName).toString() 
    paramScale, paramOffset   = self.reader.getObsScaleOffset(self.fieldToRetrieve, "scale_factor", "add_offset")
    if paramScale == None or paramOffset == None:
      self.paramScale.append(None)
      self.paramOffset.append(None)
    else:
      scales  = jarray.zeros(len(paramScale), 'f')
      offsets = jarray.zeros(len(paramOffset), 'f')
      for i in xrange(len(scales)):
         scales[i]  = paramScale[i]
         offsets[i] = paramOffset[i]
        
      self.paramScale.append(scales)
      self.paramOffset.append(offsets)
        
  def extractValues(self,dataToExtract):   
      valuesToExtract=jarray.zeros(len(dataToExtract),'f')
      if self.paramName == "Cloud Phase":
        for i in xrange(len(dataToExtract)):
          if dataToExtract[i] == 0:
            valuesToExtract[i]=0 #clear
          elif dataToExtract[i]==1 or dataToExtract[i]==5:
            valuesToExtract[i]=1 #water
          elif dataToExtract[i]==2 or dataToExtract[i]==4:
            valuesToExtract[i]=2 #ice
          elif dataToExtract[i] == 3:
            valuesToExtract[i]=3 #mixed
          elif dataToExtract[i] == 6:
            valuesToExtract[i]=4 # uncertain
          elif dataToExtract[i] == 7:
            valuesToExtract[i]=5 # missing data
      else:
        self.getScaleOffset()
        for i in xrange(len(dataToExtract)):
          if dataToExtract[i] < -9998:
            valuesToExtract[i]=Float.NaN
          else:
            valuesToExtract[i]=self.paramScale[0][0]*(dataToExtract[i]-self.paramOffset[0][0])
      
      return valuesToExtract
  
  def makeRangeTypes(self):
    self.rangeTypes = []
    paramNames=["Cloud Phase","Cloud Top Pressure","Cloud Top Temperature","Cloud Effective Radius"]
    self.imageRangeTypeDict={}
    for i in xrange(self.numObsDataSets):
        self.rangeTypes.append(RealType.getRealType(self.observationName[i]))
        self.imageRangeTypeDict[paramNames[i]]=RealType.getRealType(self.observationName[i])
    self.spectrumRangeTypeDict    = None
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None
    
  def getNativeParam(self):
    native_param="Cloud Phase"
    return  native_param

  def setDefaultParam(self):
    return self.native_param
  
  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
    if self.paramName   == "Cloud Top Pressure":
      return [0, 1000]
    elif self.paramName == "Cloud Top Temperature":
      return [180, 320]
    elif self.paramName == "Cloud Effective Radius":
      return [0, 1000]
    elif self.paramName == "Cloud Phase":
      return None
  
  def getImageRange(self, values):
    if self.paramName == "Cloud Phase":
      self.imageRangeUnit = None
    elif self.paramName == "Cloud Top Pressure":
      self.imageRangeUnit = Parser.parse("hPa")
    elif self.paramName == "Cloud Top Temperature":
      self.imageRangeUnit = Parser.parse("K")
    elif self.paramName == "Cloud Effective Radius":
      self.imageRangeUnit = Parser.parse("micron")
    return values

  def changeParameter(self, paramName):
    self.paramName = paramName
    if self.paramName == "Cloud Effective Radius":
      self.numObsElements_save = self.numObsElements
      self.numObsLines_save = self.numObsLines
      self.numObsElements = self.numObsElements_radii
      self.numObsLines    = self.numObsLines_radii
      self.ratio = self.numObsLines/self.numGeoLines
      self.trackSegments = [1, self.numObsLines]
    else:
      self.numObsElements = self.numObsElements_save
      self.numObsLines   = self.numObsLines_save
      self.ratio = self.numObsLines/self.numGeoLines
      self.trackSegments = [1, self.numObsLines]
       
  def clone(self):
    return Modis06Data(self.filename,self.reader,self.readerTable,master=self)

# MODIS Precipitable Water (for now)
class Modis07Data(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in the same directory as the MOD07 Product
    if master == None:
      if filename.find("mod07") != -1:
        geo_filename = filename.replace("mod07", "geo")
      elif filename.find("MOD07") != -1:
        geo_filename = filename.replace("MOD07", "MOD03")
      elif filename.find("MYD07") != -1:
        geo_filename = filename.replace("MYD07", "MYD03")
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"
    
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
  
  def extractValues(self,dataToExtract): 
      valuesToExtract=jarray.zeros(len(dataToExtract),'f')
      self.getScaleOffset()
      for i in xrange(len(dataToExtract)):
        if dataToExtract[i] < -9998:
          valuesToExtract[i]=Float.NaN
        else:
          valuesToExtract[i]=self.paramScale[0][0]*(dataToExtract[i]-self.paramOffset[0][0])
      
      return valuesToExtract

  def makeRangeTypes(self):
    # Create a dictionary corresponding to the names in the Level 2 Files
    self.rangeTypes = []
    paramNames=["Precipitable Water","Total Ozone","Surface Pressure","Surface Temperature"]
    self.imageRangeTypeDict={}
    for i in xrange(self.numObsDataSets):
        self.rangeTypes.append(RealType.getRealType(self.observationName[i]))
        self.imageRangeTypeDict[paramNames[i]]=RealType.getRealType(self.observationName[i])

    self.spectrumRangeTypeDict    = None
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None   
      
  def getNativeParam(self):
    native_param="Precipitable Water"
    return  native_param

  def setDefaultParam(self):
    return self.native_param

  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
    if self.paramName   == "Precipitable Water":
      return [0, 100]
    elif self.paramName == "Total Ozone":
      return [0, 500]    
    elif self.paramName == "Surface Temperature":
      return [180, 320]   
    elif self.paramName == "Surface Pressure":
      return [600, 1050]   
  
  def getImageRange(self, values):
    if self.paramName == "Precipitable Water":
      self.imageRangeUnit = Parser.parse("cm")
    elif self.paramName == "Total Ozone":
      db=BaseUnit.addBaseUnit("Dobsons","db","db")
      self.imageRangeUnit = Parser.parse(db.toString())
    elif self.paramName == "Surface Pressure":
      self.imageRangeUnit = Parser.parse("hPa")
    elif self.paramName == "Surface Temperature":
      self.imageRangeUnit = Parser.parse("K")      
    
    return values
  
  def getScaleOffset(self):
    self.paramScale   = []
    self.paramOffset  = []
    
    self.fieldToRetrieve=self.imageRangeTypeDict.get(self.paramName).toString() 
    paramScale, paramOffset   = self.reader.getObsScaleOffset(self.fieldToRetrieve, "scale_factor", "add_offset")
    if paramScale == None or paramOffset == None:
      self.paramScale.append(None)
      self.paramOffset.append(None)
    else:
      scales  = jarray.zeros(len(paramScale), 'f')
      offsets = jarray.zeros(len(paramOffset), 'f')
      for i in xrange(len(scales)):
         scales[i]  = paramScale[i]
         offsets[i] = paramOffset[i]
        
      self.paramScale.append(scales)
      self.paramOffset.append(offsets)  
      
  def clone(self):
    return Modis07Data(self.filename,self.reader,self.readerTable,master=self)
    
 # MODIS Optical Thickness Data (for now)
class Modis04Data(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in the same directory as the MOD04 Product
    if master == None:
      if filename.find("mod04") != -1:
        geo_filename = filename.replace("mod04", "geo")
      elif filename.find("MOD04") != -1:
        geo_filename = filename.replace("MOD04", "MOD03")
      elif filename.find("MYD04") != -1:
        geo_filename = filename.replace("MYD04", "MYD03")
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"
    
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
  
  def makeRangeTypes(self):
    self.rangeTypes = []
    self.rangeTypes.append(RealType.getRealType("optThickness"))    
    self.imageRangeTypeDict = {"optThickness": RealType.getRealType("optThickness")}
    self.spectrumRangeTypeDict    = None
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None
    
  def getNativeParam(self):
    native_param="optThickness"
    return  native_param

  def setDefaultParam(self):
    return self.native_param
  
  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
    if self.paramName   == "optThickness":
      return [0, 5000]  
      
  def extractValues(self,dataToExtract):   
      aot=jarray.zeros(len(dataToExtract),'f')
      for i in xrange(len(dataToExtract)):
        if dataToExtract[i] < -9998:
           #pass
           aot[i]=Float.NaN
        else:
           aot[i]=dataToExtract[i]*0.001
       
      return aot
      
  def clone(self):
    return Modis04Data(self.filename,self.reader,self.readerTable,master=self)   

# MODIS Water Vapor File
class Modis05Data(MultiSpectralData):
  def __init__(self, filename, reader, readerTable,master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
        
    #-look for MODIS geo file in the same directory as the MOD07 Product
    if master == None:
      if filename.find("mod05") != -1:
        geo_filename = filename.replace("mod05", "geo")
      elif filename.find("MOD05") != -1:
        geo_filename = filename.replace("MOD05", "MOD03")
      elif filename.find("MYD05") != -1:
        geo_filename = filename.replace("MYD05", "MYD03")
      reader_name  = reader.reader.__class__
      reader_name  = reader_name.__name__
      
      if ((filename.find("MYD") != -1) or (filename.find("a1.") != -1)):
        self.modis_platform = "Aqua"
      else:
        self.modis_platform = "Terra"
    
    else:
      self.geo_reader     = master.geo_reader
      self.geoDimOrder    = master.geoDimOrder
      self.modis_platform = master.modis_platform
  
  def extractValues(self,dataToExtract): 
      valuesToExtract=jarray.zeros(len(dataToExtract),'f')
      self.getScaleOffset()
      for i in xrange(len(dataToExtract)):
        if dataToExtract[i] < -9998:
          valuesToExtract[i]=Float.NaN
        else:
          valuesToExtract[i]=self.paramScale[0][0]*(dataToExtract[i]-self.paramOffset[0][0])
      
      return valuesToExtract

  def makeRangeTypes(self):
    # Create a dictionary corresponding to the names in the Level 2 Files
    self.rangeTypes = []
    paramNames=["Water Vapor Near Infrared","Water Vapor Correction Factors"]
    self.imageRangeTypeDict={}
    for i in xrange(self.numObsDataSets):
        self.rangeTypes.append(RealType.getRealType(self.observationName[i]))
        self.imageRangeTypeDict[paramNames[i]]=RealType.getRealType(self.observationName[i])

    self.spectrumRangeTypeDict    = None
  
  def makeSpectrumRangeType(self, paramName):
    self.spectrumRangeType = None   
      
  def getNativeParam(self):
    native_param="Water Vapor Near Infrared"
    return  native_param

  def setDefaultParam(self):
    return self.native_param

  def getObsValidRange(self): #- valid is probably a misnomer here; mostly for setRange
    if self.paramName   == "Water Vapor Near Infrared":
      return [0, 100]
    elif self.paramName == "Water Vapor Correction Factors":
      return [0, 100]    
      
  def getImageRange(self, values):
    if self.paramName == "Water Vapor Near Infrared":
      self.imageRangeUnit = Parser.parse("cm")
    elif self.paramName == "Water Vapor Correction Factors":
      self.imageRangeUnit = None
          
    return values
  
  def getScaleOffset(self):
    self.paramScale   = []
    self.paramOffset  = []
    
    self.fieldToRetrieve=self.imageRangeTypeDict.get(self.paramName).toString() 
    paramScale, paramOffset   = self.reader.getObsScaleOffset(self.fieldToRetrieve, "scale_factor", "add_offset")
    if paramScale == None or paramOffset == None:
      self.paramScale.append(None)
      self.paramOffset.append(None)
    else:
      scales  = jarray.zeros(len(paramScale), 'f')
      offsets = jarray.zeros(len(paramOffset), 'f')
      for i in xrange(len(scales)):
         scales[i]  = paramScale[i]
         offsets[i] = paramOffset[i]
        
      self.paramScale.append(scales)
      self.paramOffset.append(offsets)  
      
  def clone(self):
    return Modis05Data(self.filename,self.reader,self.readerTable,master=self)   


class WFC_calipso(MultiSpectralData):
  def __init__(self, filename, reader, readerTable, master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)

  def setDefaultParam(self):
    return "radiance"

  def makeSpectrumDomainType(self):
    micrometer = ScaledUnit(0.000001, CommonUnit.meter, "\xb5m")
    self.obsChannelsType = RealType.getRealType("wavelength", micrometer)

  #- override
  def makeTrackSegments(self):
    if self.numObsLines < 2000:
      self.numberSegments = 1
      self.trackSegments = [1, self.numObsLines]
      return

    self.trackSegments = []
    segLen = 2000
    start  = 1  #-track segment array is one's based
    self.numberSegments = int((self.numObsLines)/segLen)
    for i in xrange(self.numberSegments):
      #start = start + segLen*i
      self.trackSegments.append(start)
      self.trackSegments.append(start + (segLen-1))
      start += segLen

  #- override
  def initMissing(self, longitudes, latitudes, values):
    class Missing: #- this class must implement these methods
      def __init__(self, longitudes, latitudes, values, xlength, ylength, invert):
        self.longitudes = longitudes
        self.latitudes  = latitudes
        self.xlength    = xlength
        self.ylength    = ylength
        self.invert     = invert

      def dataToSegment(self, line, element):
        return line, element

      def removeMissing(self, values):
        for i in xrange(len(values)):
          if values[i] > 30000.0:
            values[i] = Float.NaN
        return values

      def getLengths(self):
        return self.xlength, self.ylength

    m = Missing(longitudes, latitudes, values, self.xlength, self.ylength, self.invert)
    return m

  def clone(self):
    return WFC_calipso(self.filename,self.reader,self.readerTable,master=self)

class AMSU_A(MultiSpectralData):
  def __init__(self, filename, reader, readerTable, master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)

  def getNativeParam(self):
    native_param="brightnessTemp"
    return native_param

  def setDefaultParam(self):
    return self.native_param

  def getObservationChannels(self):
    self.fileIndexes = None
    self.observationChannels = [23.8,31.4,50.3,52.8,53.596,54.4,54.94,55.5,57.29,57.39,57.49,57.59,57.69,57.79,89.0]
    self.bandDict = {}
    self.band_numbers = ['1','2','3','4','5','6','7','8','9','10','11','12','13','14','15']
    self.fileIndexes = QuickSort.sort(self.observationChannels)
    cnt = 0
    for i in xrange(len(self.observationChannels)):
      band_number = self.band_numbers[i]
      center_wn = self.observationChannels[i]
      self.bandDict.update({band_number : [cnt,center_wn]})
      self.bandDict.update({center_wn : band_number})
      cnt += 1
                                                                                                                               
    return self.observationChannels

  def bandNameToValue(self, text):
    return (self.bandDict[text])[1]
                                                                                                                               
  def bandValueToName(self, value):
    idx = self.observationDomain.valueToIndex([[value]])[0]
    return self.band_numbers[idx]



class AVHRR_seaspace(MultiSpectralData):
  def __init__(self, filename, reader, readerTable, master=None):
    MultiSpectralData.__init__(self, filename, reader, readerTable, master)
    self.instrumentName = "AVHRR"
    if master == None:
      self.instrumentName = "AVHRR"
      self.getScaleOffset()
    else:
      self.day = master.day
      self.scale_s = master.scale_s
      self.offset_s = master.offset_s

  def getObservationUnit(self):
    self.observationUnit.append(Parser.parse("C"))

  def getNativeParam(self):
    native_param="brightnessTemp"
    return  native_param

  def setDefaultParam(self):
    return self.native_param

  def getDayNight(self):
    day = false
    top_day = false
    bot_day = false

    any_day = false
    any_night = false
    for i in xrange(self.numObsLines/2):
      start  =  [0,i]
      count  =  [self.numObsElements/10, 1]
      stride =  [10,1]
      nstart =  []
      ncount =  []
      nstride = []
                                                                                                                                            
      for i in xrange(len(start)):
        nstart.append(start[self.geoDimOrder[i]])
        ncount.append(count[self.geoDimOrder[i]])
        nstride.append(stride[self.geoDimOrder[i]])

      vals = self.reader.getValues("sun_zenith",nstart,ncount,nstride)
      for j in xrange(len(vals)):
        zenith = vals[j]
        if (zenith > -88.0 and zenith < 88.0):
          any_day = true
        elif (zenith > -180.0 and zenith < 180.0):
          any_night = true

      if (any_day == true or any_night == true):
        break
      
    if (any_day == true and any_night == true):
      print "day/night mix:  assume channel 3A"
      return true
    elif (any_day == true):
      top_day = true
    else:
      top_day = false

    any_day = false
    any_night = false
    for i in xrange(self.numObsLines/2):
      start  =  [0,(self.numObsLines-1)-i]
      count  =  [self.numObsElements/10, 1]
      stride =  [10,1]
      nstart =  []
      ncount =  []
      nstride = []
                                                                                                                                            
      for i in xrange(len(start)):
        nstart.append(start[self.geoDimOrder[i]])
        ncount.append(count[self.geoDimOrder[i]])
        nstride.append(stride[self.geoDimOrder[i]])
                                                                                                                                            
      vals = self.reader.getValues("sun_zenith",nstart,ncount,nstride)
      for j in xrange(len(vals)):
        zenith = vals[j]
        if (zenith > -88.0 and zenith < 88.0):
          any_day = true
        elif (zenith > -180 and zenith < 180):
          any_night = true

      if (any_day == true or any_night == true):
        break
                                                                                                                                            
    if (any_day == true and any_night == true):
      print "day/night mix:  assume channel 3A"
      return true
    elif (any_day == true):
      bot_day = true
    else:
      bot_day = false

    if (top_day == true):
      if (bot_day == true):
        day = true
      else:
        print "day/night mix:  assume channel 3A"
        return true

    if (top_day == false):
      if (bot_day == false):
        day = false
      else:
        print "day/night mix:  assume channel 3A"
        return true

    return day

  def getObservationChannels(self):
    self.day = self.getDayNight()
    channels = []
    self.band_numbers = []
    ff       = open(hydra_path+'/ancillary/avhrr/avhrr_3', 'r')
    lines    = ff.readlines()
    nlines   = len(lines)
    self.bandDict = {}
    cnt = 0
    for i in xrange(nlines):
      if lines[i].startswith('!'):
        pass
      else:
        fields      = lines[i].split()
        band_number = fields[0]
        center_wn   = fields[1]
        low_wn      = fields[2]
        hi_wn       = fields[3]
        center_wn   = float(center_wn)
        if band_number == '3A':
          if self.day == false:
            continue
        if band_number == '3B':
          if self.day == true:
            continue
        self.band_numbers.append(band_number)
        channels.append(center_wn)
        self.bandDict.update({band_number : [cnt,center_wn]})
        self.bandDict.update({center_wn : band_number})
        cnt += 1

    self.fileIndexes = QuickSort.sort(channels)

    return channels

  def getScaleOffset(self):
    self.scale_s   = []
    self.offset_s  = []

    for i in xrange(self.numObsDataSets):
      scale, offset   = self.reader.getObsScaleOffset(self.observationName[i], "scale_factor", "add_offset")
      if scale == None or offset == None:
        self.scale.append(None)
        self.offset.append(None)
      else:
        scales  = jarray.zeros(len(scale), 'f')
        offsets = jarray.zeros(len(offset), 'f')
        for i in xrange(len(scales)):
          scales[i]  = scale[i]
          offsets[i] = offset[i]
        self.scale_s.append(scales)
        self.offset_s.append(offsets)

  def getSpectrumRange(self, values, paramName):
    if paramName == "brightnessTemp":
      scales  = jarray.zeros(self.numObsChannels, 'f')
      offsets = jarray.zeros(self.numObsChannels, 'f')

      cnt = 0
      for i in xrange(self.numObsDataSets):
        System.arraycopy(self.scale_s[i], 0, scales, cnt, self.numChannelsPerDataSet[i])
        System.arraycopy(self.offset_s[i], 0, offsets, cnt, self.numChannelsPerDataSet[i])
        cnt += self.numChannelsPerDataSet[i]

      vals = Helper.unpack(values, scales, offsets)
      vals1_2  = jarray.zeros(len(values), 'f')
      vals3_5  = jarray.zeros(len(values), 'f')

      for i in xrange(len(values)):
        if i <= 1:
          vals1_2[i] = vals[i]
          vals3_5[i] = Float.NaN
        else:
          vals1_2[i] = Float.NaN
          vals3_5[i] = vals[i]

      return [vals1_2, vals3_5]

  def getImageRange(self, values, paramName, dataset_index, index):
    if paramName == "radiance":
      unit = Parser.parse("W/m^2/micrometer/steradian")
      self.imageRangeUnit = ScaledUnit(1, unit, "W/m^2/\xb5m/ster")
      return Helper.unpack(values, self.scaler_s[dataset_index][index], self.offset_s[dataset_index][index])
    elif paramName == "brightnessTemp":
      self.imageRangeUnit = Parser.parse("C")
      return Helper.unpack(values, self.scale_s[dataset_index][index], self.offset_s[dataset_index][index])

  def makeRangeTypes(self):
    self.spectrumRangeTypeDict = {
      "radiance":
        RealTupleType(RealType.getRealType("Radiance_1-2"), RealType.getRealType("Radiance_3-5")),
      "brightnessTemp":
        RealTupleType(RealType.getRealType("Albedo*100"), RealType.getRealType("BrightnessTemp"))}
                                                                                                                                            
    self.imageRangeTypeDict = {
      "radiance": RealType.getRealType("Radiance"),
      "brightnessTemp": RealType.getRealType("Albedo/BT")}
                                                                                                                                            
    self.rangeTypes = []
    self.rangeTypes.append(RealType.getRealType("Radiance_1-2"))
    self.rangeTypes.append(RealType.getRealType("Radiance_3-5"))
    self.rangeTypes.append(RealType.getRealType("Albedo*100"))
    self.rangeTypes.append(RealType.getRealType("BrightnessTemp"))
                                                                                                                                            
  def makeSpectrumDomainType(self):
    micrometer = ScaledUnit(0.000001, CommonUnit.meter, "\xb5m")
    self.obsChannelsType = RealType.getRealType("wavelength", micrometer)
                                                                                                                                            
  def makeSpectrumRangeType(self, paramName):
    if paramName == "radiance":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[0], self.rangeTypes[1])
    elif paramName == "brightnessTemp":
      self.spectrumRangeType = RealTupleType(self.rangeTypes[2], self.rangeTypes[3])
                                                                                                                                            
  def makeImageRangeType(self, paramName):
    if paramName == "radiance":
      self.imageRangeType = RealType.getRealType("radiance")
    elif paramName == "brightnessTemp":
      self.imageRangeType = RealType.getRealType("brightnessTemp")
                                                                                                                                            
  def bandNameToValue(self, text):
    return self.observationChannels[(self.bandDict[text])[0]]
                                                                                                                                            
  def bandValueToName(self, value):
    idx = self.observationDomain.valueToIndex([[value]])[0]
    return self.band_numbers[idx]

  def getObsValidRange(self):
    if self.paramName   == "radiance":
      return [[0,500],[0, 10]]
    elif self.paramName == "brightnessTemp":
      return [[0,100],[-90, 50]]

  def clone(self):
    return AVHRR_seaspace(self.filename,self.reader,self.readerTable,master=self)
